// VinciaMGInterface.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "VinciaMGInterface.h"

using namespace Pythia8;

// Function to initialize the MadGraph coupling common blocks, required
// by the MG matrix elements. Adapted from couplings.f in the MG 
// Models/sm/ directory.

namespace Vincia {

  // Wrappers for the MadGraph F77 matrix elements
  
  extern "C" {
    
    // Wrappers for common blocks, ported from coupl.inc
    extern struct {
      double          g;	
      complex<double> gg[2];	
    } coupl_qcd_;

    extern struct {
      double          hmass, wmass, zmass, amass, tmass, bmass, lmass, cmass;
    } coupl_mass_;
    
    extern struct {
      double          hwidth, wwidth, zwidth, twidth, lwidth, awidth;
    } coupl_width_;
    
    extern struct {
      complex<double> gal[2], gad[2], gau[2], gwf[2], gzn[2], gzl[2], gzd[2], gzu[2];
      double          gw, gwwa, gwwz;      
    } coupl_gauge_;
    
    extern struct {
      complex<double> gwfud[2],  gwfcd[2], gwftd[2],
	gwfus[2],  gwfcs[2], gwfts[2],
	gwfub[2],  gwfcb[2], gwftb[2];
    } coupl_ckm_;
    
    // Uncomment for MadGraph with diagonal CKM:
    /*
    extern struct {
      complex<double> gwfc[2],  gwfs[2], gwfm[2];    
    } coupl_ckm_;
    */

    extern struct {
      complex<double> gwwh, gzzh, gwwhh, gzzhh, ghhh, ghhhh;
    } coupl_scal_;
    
    extern struct {
      complex<double> ghtop[2], ghbot[2], ghtau[2], ghcha[2];
    } coupl_yuk_;
    
    // Wrappers for MG decay matrix elements
    // Born
    extern void sz2bbx_(int *cMode, double P1[3][4],double ANS[1]);
    extern void sw2bcx_(int *cMode, double P1[3][4],double ANS[1]);
    // 1st Order
    extern void sz2bgbx_(int *cMode, double P1[4][4],double ANS[1]);
    // 2nd Order
    extern void sz2bggbx_(int *cMode, double P1[5][4],double ANS[1]);
    extern void sz2bcxcbx_(int *cMode, double P1[5][4],double ANS[1]);
    // 3rd Order
    extern void sz2bgggbx_(int *cMode, double P1[6][4],double ANS[1]);
    extern void sz2bcxcgbx_(int *cMode, double P1[6][4],double ANS[1]);
    extern void sz2bgcxcbx_(int *cMode, double P1[6][4],double ANS[1]);
    
    // 4th Order
    extern void sz2bggggbx_(int *cMode, double P1[7][4],double ANS[1]);
    extern void sz2bcxcggbx_(int *cMode, double P1[7][4],double ANS[1]);
    extern void sz2bgcxcgbx_(int *cMode, double P1[7][4],double ANS[1]);
    extern void sz2bggcxcbx_(int *cMode, double P1[7][4],double ANS[1]);
    extern void sz2bcxcuxubx_(int *cMode, double P1[7][4],double ANS[1]);
    extern void sz2buxucxcbx_(int *cMode, double P1[7][4],double ANS[1]);
  }

  //*********

  // Initialize MadGraph

  bool MGInterface::init() {
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // If matching not requested, just return
    if (settingsPtr->mode("VinciaMatching:order") <= 0) return false;

    // Verbose output
    if (verbose >= 2) printOut("MGInterface::init","...");

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("MGInterface::init","Cannot initialize, pointers not set.");
      return false;      
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting MG masses...");

    // Set who is massive
    isMassiveS   = settingsPtr->flag("Vincia:isMassiveS");
    isMassiveC   = settingsPtr->flag("Vincia:isMassiveC");
    isMassiveB   = settingsPtr->flag("Vincia:isMassiveB");
    isMassiveTau = settingsPtr->flag("Vincia:isMassiveTau");

    // Set masses and widths (initialized through Pythia)
    coupl_mass_.cmass   = (isMassiveC ? particleDataPtr->m0(4) : 0.0);
    coupl_mass_.bmass   = (isMassiveB ? particleDataPtr->m0(5) : 0.0);
    coupl_mass_.tmass   = particleDataPtr->m0(6);
    coupl_mass_.lmass   = (isMassiveTau ? particleDataPtr->m0(15) : 0.0);
    coupl_mass_.amass   = particleDataPtr->m0(22);
    coupl_mass_.zmass   = particleDataPtr->m0(23);
    coupl_mass_.wmass   = particleDataPtr->m0(24);
    coupl_mass_.hmass   = particleDataPtr->m0(25);
    coupl_width_.twidth = particleDataPtr->mWidth(6);
    coupl_width_.lwidth = (isMassiveTau ? particleDataPtr->mWidth(15) : 0.0);
    coupl_width_.awidth = particleDataPtr->mWidth(22);
    coupl_width_.zwidth = particleDataPtr->mWidth(23);
    coupl_width_.wwidth = particleDataPtr->mWidth(24);
    coupl_width_.hwidth = particleDataPtr->mWidth(25);

    // Useful shorthands
    const double Fourpi = 4.0 * M_PI;
    const double Zero   = 0.0;
    // Here using alpha(mZ) (as in SLHA) rather than alpha(0)
    // Check consistency when doing full matrix elements.
    const double alpha  = settingsPtr->parm("StandardModel:alphaEMmZ");
    // Use thetabar for VFF couplings? (as in Pythia)
    // Check consistency when doing full matrix elements.
    const double sin2w  = couplingsPtr->sin2thetaW();
    const double cw     = sqrt( 1.0 - sin2w );
    const double ee2    = alpha * Fourpi;
    const double sw     = sqrt( sin2w );
    const double ee     = sqrt( ee2 );
    const double ez     = ee/(sw*cw);
    const double ey     = ee*(sw/cw);
    const double sc2    = sin2w*( 1.0 - sin2w );
    const double v      = 2.0*coupl_mass_.wmass*sw/ee;
    //    const double lambda = pow2(coupl_mass_.hmass) / (2.0 * pow2(v));

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting MG couplings...");

    // Set QCD g = 1 to keep factored out (=> alphaS = 1/sqrt(4pi))
    coupl_qcd_.g      = 1.0;
    coupl_qcd_.gg[0]  = -coupl_qcd_.g;
    coupl_qcd_.gg[1]  = -coupl_qcd_.g;

    // Set Vector boson couplings
    coupl_gauge_.gw   = ee/sw;
    coupl_gauge_.gwwa = ee;
    coupl_gauge_.gwwz = ee*cw/sw;
    
    // Set gauge & higgs boson couplings
    coupl_scal_.gwwh  = complex<double>( ee2/sin2w*0.5*v, Zero );
    coupl_scal_.gzzh  = complex<double>( ee2/sc2*0.5*v  , Zero );
    coupl_scal_.ghhh  = complex<double>( -pow2(coupl_mass_.hmass)/v*3.0, Zero );
    coupl_scal_.gwwhh = complex<double>( ee2/sin2w*0.5  , Zero );
    coupl_scal_.gzzhh = complex<double>( ee2/sc2*0.5    , Zero );
    coupl_scal_.ghhhh = coupl_scal_.ghhh/v;

    // Set fermion-fermion-vector couplings
    coupl_gauge_.gal[0] = complex<double>(  ee          , Zero );
    coupl_gauge_.gal[1] = complex<double>(  ee          , Zero );
    coupl_gauge_.gau[0] = complex<double>( -ee*2./3.    , Zero );
    coupl_gauge_.gau[1] = complex<double>( -ee*2./3.    , Zero );
    coupl_gauge_.gad[0] = complex<double>(  ee/3.       , Zero );
    coupl_gauge_.gad[1] = complex<double>(  ee/3.       , Zero );

    coupl_gauge_.gwf[0] = complex<double>( -ee/sqrt(2.*sin2w) , Zero );
    coupl_gauge_.gwf[1] = complex<double>(  Zero              , Zero );

    coupl_gauge_.gzn[0] = complex<double>( -ez*0.5                  , Zero );
    coupl_gauge_.gzn[1] = complex<double>(  Zero                    , Zero );
    coupl_gauge_.gzl[0] = complex<double>( -ez*(-0.5 + sin2w)       , Zero );
    coupl_gauge_.gzl[1] = complex<double>( -ey                      , Zero );
    coupl_gauge_.gzu[0] = complex<double>( -ez*( 0.5 - sin2w*2./3.) , Zero );
    coupl_gauge_.gzu[1] = complex<double>(  ey*2./3.                , Zero );
    coupl_gauge_.gzd[0] = complex<double>( -ez*(-0.5 + sin2w/3.)    , Zero );
    coupl_gauge_.gzd[1] = complex<double>( -ey/3.                   , Zero );

    // Should also fill in Yukawa couplings and CKM, but more tricky due
    // to running, so take up again later.

    // Verbose output
    if (verbose >= 3) printOut("MGInterface::init","   setting Color Depth...");

    // Set color depth
    colorDepth = (settingsPtr->flag("VinciaMatching:fullColor")) ? 1 : 0;

    return true;
  }
  
  //*********

  // Wrapper to set values in the MG common blocks

  void MGInterface::setParm(string txt, complex<double> val) {

    // Check if initialized
    if (! isInit) {
      printOut("MGInterface::setParm","MGInterface not initialized.");
    }
    // Convert to all lower-case
    for (unsigned int i=0;i<txt.length();i++) txt[i]=tolower(txt[i]);
    if (txt == "zmass" || txt == "mz") coupl_mass_.zmass   = val.real();
    else if (txt == "wmass" || txt == "mw") coupl_mass_.wmass   = val.real();
    else if (txt == "bmass" || txt == "mb") coupl_mass_.bmass   = val.real();
    // Z couplings
    else if (txt == "gzu0") coupl_gauge_.gzu[0]=val;
    else if (txt == "gzu1") coupl_gauge_.gzu[1]=val;
    // W couplings
    else if (txt.find("gwf") != string::npos) {
      if (txt == "gwf0") coupl_gauge_.gwf[0]=val;
      else if (txt == "gwf1") coupl_gauge_.gwf[1]=val;
      // CKM
      else if (txt == "gwfud0") coupl_ckm_.gwfud[0]=val;
      else if (txt == "gwfud1") coupl_ckm_.gwfud[1]=val;
      else if (txt == "gwfcd0") coupl_ckm_.gwfcd[0]=val;
      else if (txt == "gwfcd1") coupl_ckm_.gwfcd[1]=val;
      else if (txt == "gwftd0") coupl_ckm_.gwftd[0]=val;
      else if (txt == "gwftd1") coupl_ckm_.gwftd[1]=val;
      else if (txt == "gwfus0") coupl_ckm_.gwfus[0]=val;
      else if (txt == "gwfus1") coupl_ckm_.gwfus[1]=val;
      else if (txt == "gwfcs0") coupl_ckm_.gwfcs[0]=val;
      else if (txt == "gwfcs1") coupl_ckm_.gwfcs[1]=val;
      else if (txt == "gwfts0") coupl_ckm_.gwfts[0]=val;
      else if (txt == "gwfts1") coupl_ckm_.gwfts[1]=val;
      else if (txt == "gwfub0") coupl_ckm_.gwfub[0]=val;
      else if (txt == "gwfub1") coupl_ckm_.gwfub[1]=val;
      else if (txt == "gwfcb0") coupl_ckm_.gwfcb[0]=val;
      else if (txt == "gwfcb1") coupl_ckm_.gwfcb[1]=val;
      else if (txt == "gwftb0") coupl_ckm_.gwftb[0]=val;
      else if (txt == "gwftb1") coupl_ckm_.gwftb[1]=val;
    }
    else {
      printOut("MGInterface::setParm","Unknown parameter "+txt);
    }

  }

  void MGInterface::setParm(string txt, double val) {

    complex<double> valC(val,0.0);
    setParm(txt, valC);

  }

  //*********

  // Wrapper to return MadGraph Matrix Elements (squared)
  // Input:
  //    nIn   = number of particles that are incoming (1 or 2)
  //    id    = vector of particle IDs. The incoming particle(s) MUST be the 
  //            first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 
  //    p     = vector of particle momenta. The incoming particle(s) MUST be 
  //            the first nIn entries of this vector. The outgoing particles
  //            MUST be arranged in color order thereafter. 

  // Return:   value of Matrix Element squared for this phase-space point

  double MGInterface::ME2(int nIn, vector<int> id, vector<Vec4> p) {

    // Check if initialized
    if (! isInit) {
      printOut("MGInterface::setParm","MGInterface not initialized.");
    }
    
    // Initialize ME weight with dummy value.
    const  int nMom  = p.size();
    const  int nOut  = nMom-nIn;
    double MGME2 = -1.;

    // Take out polarization averages
    int nPolarizations = 1;
    
    // Only DECAYS implemented so far
    if (nIn != 1 || int(id.size()) != nMom || nOut <= 1) return MGME2;

    //------------------------------------------------------------------------
    // Decays: determine if we know the relevant Lorentz/Color structure
    int id0 = 0;
    // Color singlets
    if (particleDataPtr->colType(id[0]) == 0) {
      // Spin-1 (e.g., W, Z, Z', ...)
      if (particleDataPtr->spinType(id[0]) == 3) id0 = 23;
    }

    // If not known, return.
    if (id0 == 0) return -1;

    // Check if using massive matrix elements
    bool massless=true;
    for (unsigned int i=1;i<id.size(); i++) 
      if (vinComPtr->isMassive(id[i])) massless = false;
    
    // Test (and enforce) massless condition
    if (massless) {
      for (unsigned int i=1;i<p.size(); i++) {
	double mu = p[i].m2Calc()/p[0].m2Calc();
	if (mu > TINY) {
	  if (verbose >= 3) {
	    printOut("VinciaMGInterface::ME2",
		     "Warning: massless ME requested, momentum vector "
		     +num2str(int(i),2)+" m ="+num2str(p[i].mCalc()));
	  }
	  // Enforcement of massless condition not yet implemented ...
	  massless = false;
	}
      }
    }

    //------------------------------------------------------------------------
    // Color-singlet spin-1 particles (from Z and W decays)
    if (id0 == 23) {

      // Polarizations
      nPolarizations = 3;

      // Ensure ME evaluated on shell
      const double mSys = p[0].mCalc();
      coupl_mass_.zmass = mSys;
      coupl_mass_.wmass = mSys;
      
      // Z->2
      if (nOut == 2) {

	// Convert momenta to MadGraph format
	double PNEW[3][4];
	for (int i=0; i<int(nMom); i++) Vec4toMG(p[i],PNEW[i]);

	// Evaluate Matrix Element
        if (abs(id[1]) == abs(id[2])) {
          coupl_mass_.bmass = p[1].mCalc();
          // Massive (equal masses, from Z->bbbar)
	  sz2bbx_(&colorDepth,PNEW,&MGME2);
        }
        else if (abs(id[1]) != abs(id[2])) {
	  // Massive (unequal masses, from W->cbbar)
	  coupl_mass_.bmass = p[1].mCalc();
	  coupl_mass_.cmass = p[2].mCalc();
	  sw2bcx_(&colorDepth,PNEW,&MGME2);
	}
	if (verbose >= 8) {
	  printOut("VinciaMGInterface::ME2","Getting M2 with colDepth = "
		   +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	}

	// TMP! 
	/*
	cout<<" nMom = "<<nMom<<" nOut = "<<nOut<<" nIn = "<<nIn<<endl;
	for (int i=0; i<p.size(); i++) cout << "  "<<i<<" "<<p[i].px() <<" "<<p[i].py() << " "
				      << p[i].pz() <<" "<<p[i].e()<<endl;
	for (int i=0; i<p.size(); i++) cout << "  "<<i<<" "<<PNEW[i][1] <<" "<<PNEW[i][2] << " "
				      << PNEW[i][3] <<" "<<PNEW[i][0]<<endl;
	printOut("VinciaMGInterface::ME2","Getting Z->2 ME "<<MGME2<<" with CD = "<<colorDepth<<endl;
	*/

      }

      // Z->3
      else if (nOut == 3) {

	// Convert momenta to MadGraph format
	double PNEW[4][4];
	for (int i=0; i<int(nMom); i++) Vec4toMG(p[i],PNEW[i]);

	// Evaluate Matrix Element
	if (abs(id[1]) == abs(id[3])) {
	  // Massive (equal masses, from Z->bgbbar)
	  coupl_mass_.bmass = p[1].mCalc();
	  sz2bgbx_(&colorDepth,PNEW,&MGME2);
       	} else if (abs(id[1]) != abs(id[3])) {
	  // Massive (unequal masses, from W->cgbbar)
	  coupl_mass_.cmass = p[1].mCalc();
	  coupl_mass_.bmass = p[3].mCalc();
	  // Not implemented yet
	  sz2bgbx_(&colorDepth,PNEW,&MGME2);
	}

	if (verbose >= 8) {
	  printOut("VinciaMGInterface::ME2","Getting M3 with colDepth = "
		   +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	}

      }

      // Higher than 3, need to distinguish how many g's and how many q's
      else {
	int cSum = 0;
        // colType according to Pythia documentation:
        // the colour type, with 0 uncoloured, 1 triplet, 
        // -1 antitriplet and 2 octet
	for (int i=1; i<nMom; i++) cSum += particleDataPtr->colType(id[i]);

	// Z->4
	if (nOut == 4) {

	  // Convert momenta to MadGraph format
	  double PNEW[5][4];
	  for (int i=0; i<int(nMom); i++) Vec4toMG(p[i],PNEW[i]);

	  // Z->qggqbar
	  if (cSum == 4) { 
	    
	    // Evaluate Matrix Element	    
	    if (abs(id[1]) == abs(id[4])) {

	      // Massive (equal masses)
	      coupl_mass_.bmass = p[1].mCalc();
	      sz2bggbx_(&colorDepth,PNEW,&MGME2);

	    }

	    if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M4 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	    }
	  } 

	  // Z->qq'barq'qbar
	  else {
	    // the matrix element has no up-type z coupling
	    coupl_mass_.bmass = p[1].mCalc();
            coupl_mass_.cmass = p[2].mCalc();
            sz2bcxcbx_(&colorDepth,PNEW,&MGME2);

	    if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M4 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	    }
	  }
	}

	// Z->5 
	else if (nOut == 5) {

	  // Convert momenta to MadGraph format
	  double PNEW[6][4];
	  for (int i=0; i<int(nMom); i++) Vec4toMG(p[i],PNEW[i]);

	  // Z->qgggqbar
	  if (cSum == 6) {

	    if (abs(id[1]) == abs(id[5])) {
              // MR 7.3.2011: set b mass to mass of quark 1
              coupl_mass_.bmass = p[1].mCalc();
	      sz2bgggbx_(&colorDepth,PNEW,&MGME2);
	    }
	  }

	  // Z->qgqbar'q'qbar and Z->qqbar'q'gqbar
	  else {
            // Z->qgqbar'q'qbar
	    if (id[2] == 21) {
              coupl_mass_.bmass = p[1].mCalc();
              coupl_mass_.cmass = p[3].mCalc();
	      sz2bgcxcbx_(&colorDepth,PNEW,&MGME2);
	    }
	    // Z->qqbar'q'gqbar
	   else if (id[4] == 21) {
              coupl_mass_.bmass = p[1].mCalc();
              coupl_mass_.cmass = p[2].mCalc();
	      sz2bcxcgbx_(&colorDepth,PNEW,&MGME2);
	    }
          }
          if (verbose >= 8) {
	      printOut("VinciaMGInterface::ME2","Getting M5 with colDepth = "
		       +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
          }
	} // end nOut == 5

	// Z->6
	else if (nOut == 6) {

	  // Convert momenta to MadGraph format
	  double PNEW[7][4];
	  for (int i=0; i<int(nMom); i++) Vec4toMG(p[i],PNEW[i]);

	  // Z->qggggqbar
	  if (cSum == 8 && (id[1]+id[6]==0) ) {
            // set quark mass
	    coupl_mass_.bmass = p[1].mCalc();
            sz2bggggbx_(&colorDepth,PNEW,&MGME2);
	  }
          // Z->q qbar' q' g g qbar
          else if (cSum == 4 && (id[1]+id[6]==0) && (id[2]+id[3]==0)) {
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = p[1].mCalc();
            coupl_mass_.cmass = p[2].mCalc();
            sz2bcxcggbx_(&colorDepth,PNEW,&MGME2);          
          }
          // Z->g q qbar' q' g qbar
          else if (cSum == 4 && (id[1]+id[6]==0) && (id[3]+id[4]==0)) {
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = p[1].mCalc();
            coupl_mass_.cmass = p[3].mCalc();
            sz2bgcxcgbx_(&colorDepth,PNEW,&MGME2);
          }
          // Z->q g g qbar' q' qbar
          else if (cSum == 4 && (id[1]+id[6]==0) && (id[4]+id[5]==0)) {
            // the matrix element has no Z-u-type coupling
            // set b-mass to that of primary quark, c to secondary
            coupl_mass_.bmass = p[1].mCalc();
            coupl_mass_.cmass = p[4].mCalc();
            sz2bggcxcbx_(&colorDepth,PNEW,&MGME2);
          }
          // Z->q q' qbar' q'' qbar'' qbar
          // the MG interface does not allow at present to make a 
          // light quark massive, therefore one of the secondary quarks 
          // involved has to be massless
          else if (cSum == 0 && id[1]+id[6]==0 && id[2]+id[3]==0 && 
            id[4]+id[5]==0
            ) {
            // the matrix element has no Z-u-type coupling
            if (!vinComPtr->isMassive(id[4])) {
              coupl_mass_.bmass = p[1].mCalc();
              coupl_mass_.cmass = p[2].mCalc();
              sz2bcxcuxubx_(&colorDepth,PNEW,&MGME2);
            }
            else if (!vinComPtr->isMassive(id[2])) {
              coupl_mass_.bmass = p[1].mCalc();
              coupl_mass_.cmass = p[4].mCalc();
              sz2buxucxcbx_(&colorDepth,PNEW,&MGME2);
            }
            // case of both secondary quarks massive not implemented
          }


	  
	  if (verbose >= 8) {
	    printOut("VinciaMGInterface::ME2","Getting M6 with colDepth = "
		     +num2str(colorDepth,1)+" ME = "+num2str(MGME2));
	  }
	}
	
      }
      
    }

    if (isnan(MGME2) && verbose >= 1) {
      printOut("VinciaMGInterface::ME2","Error! ME2 = nan. Printing momenta:");
      for (int i=0; i<int(p.size()); i++) {
	cout<<"   p"<<i<<p[i];
      }
    }

    // Return |ME|^2 value (or -1 if no ME found)
    if (abs(MGME2+1) < TINY) return -1.;
    else return MGME2/nPolarizations;

  }

}
