// VinciaAntenna.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2011 P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the VinciaAntenna header file
#include "VinciaAntenna.h"

using namespace Pythia8;

namespace Vincia {
  
  //**********************************************************************
  
  // The Antenna class and its offspring
  
  bool Antenna::init(bool checkAntenna) {

    // Check that pointers initialized
    if (! isInitPtr) {
      printOut("Antenna::init","Cannot initialize, pointers not set.");
      return false;
    }
    
    // Innocent until proven guilty
    isInit=true;

    // Set verbose level
    setParm(verbose,"verbose");

    // Check if switched off
    if ( ! settingsPtr->flag(name())) {
        clearC();
	isOnSav = false;
	return true;
    } else {
      isOnSav = true;
    }

    // Set whether collinear and finite terms to be used in evolution
    isInit=isInit&&setParm(useCollinearTerms,"useCollinearTerms");
    isInit=isInit&&setParm(useFiniteTerms,"useFiniteTerms");
    // MR 2.3.2011: Set whether massive terms are used    
    isInit=isInit&&setParm(useMassTerms,"useMassTerms");
    // Read octetPartitioning parameter
    isInit=isInit&&setParm(octetPartitioning,"octetPartitioning");
    // Initialize parameter to use when checking for dead zones
    isInit=isInit&&setParm(deadZoneAvoidance,"deadZoneAvoidance");
    
    // Check charge conservation
    double sumQ=0.0;
    for (unsigned int ic=0;ic<idOld.size();ic++) 
      sumQ+=particleDataPtr->charge(idOld[ic]);
    for (unsigned int ic=0;ic<idNew.size();ic++) 
      sumQ-=particleDataPtr->charge(idNew[ic]);
    if (abs(sumQ) > TINY) {
      if (verbose >= 1)
	printErr(name()+":init","antenna does not conserve charge");
      cout<<"   sumQ = "<<sumQ<<" Antenna ID :"<<idOld[1]<<" "
	  <<idOld[2]<<" -> "
	  <<idNew[1]<<" "<<idNew[2]<<" "<<idNew[3]<<endl;
      return false; 
    }

    // Set coefficients of singular terms of antenna
    // Base class implementation corresponds to Eikonal
    setC(-1,-1,2.0);
    setC(-1, 0,2.0);
    setC(-1, 0,2.0);

    // Read in antenna function finite terms (zero if switched off)        
    readC();  

    // Check that antenna function is positive definite 
    // and free of dead zones, if not rescale the finite terms
    if (checkAntenna) {
      bool initCheckPassed = checkInit(gettestMasses());
      if (!initCheckPassed) {
        return false;
      }
    }

    // Initialize evolution settings
    isInit=isInit&&setParm(evolutionType,"evolutionType"); 
    isInit=isInit&&setParm(kineMapType,"kineMapType");
    
    // Set chargeFactor (i.e., CA, CF, TR, etc.)
    isInit=isInit&&setParm(chargeFactor,"chargeFactor");
        
    // Return true or false, depending on how checks went.
    return isInit;
  }
  
  // the check for positivity and absence of dead zones called 
  // in init()
  bool Antenna::checkInit(vector< vector<double> > testMasses) {
    bool changedForPositivity = false;
    bool changedForNoDead = false;
    int nPointsCheck = 1000;
    
    for (unsigned int i = 0; i < testMasses.size(); ++i) {
      double mu1 = testMasses[i][0];
      double mu2 = testMasses[i][1];
      double mu3 = testMasses[i][2];
      if (verbose >= 4) {
        printOut(name() + ":init", 
                 "checking with reduced masses " +
                 num2str( mu1 ) + ", " +
                 num2str( mu2 ) + ", " +
                 num2str( mu3 ) + ".");
      }
      // if all the positivity tests passed so far, call
      // IsPositive without instructing it to repair
      if (not changedForPositivity) {
        changedForPositivity = not IsPositive(nPointsCheck, 
                                              mu1, mu2, mu3, false);
      }
      // if a positivity test failed up to here, call IsPositive
      // instructing it to change the parameters if needed
      if (changedForPositivity) {
        if (not IsPositive(nPointsCheck,mu1,mu2,mu3,true)) {
          return false;
        };
      }
      // the analogue for dead zones
      if (not changedForNoDead) {
        changedForNoDead = not IsFreeOfDeadZones(
                                                 nPointsCheck, 
                                                 mu1, mu2, mu3, false);
      }
      if (changedForNoDead) {
        if (not IsFreeOfDeadZones(nPointsCheck, 
                                  mu1, mu2, mu3, true) ) {
          return false;
        };
      }
      
    }
    
    // give a message if coefficients were changed and the 
    // verbosity is high enough
    
    if (changedForPositivity) {
      if (verbose >= 1)
        printErr(name()+":init","warning: Finite terms changed to ensure positivity");
      listCoefficients();
    } else if (changedForNoDead) {
      if (verbose >= 1)
        printErr(name()+":init","warning: Finite terms changed to ensure absence of dead zones");
      listCoefficients();
    } else {
      if (verbose >= 4)
        printOut(name()+":init","ok: positive definite, no dead zones");
    }
    // if the function arrives here, things are OK
    return true;
  }
  
  bool Antenna::setParm(bool& par, string parName) {
  bool found=true;
  if (settingsPtr->isFlag(name()+":"+parName)) 
    par = settingsPtr->flag(name()+":"+parName);
  else if (settingsPtr->isFlag("VinciaShower:"+parName)) 
    par = settingsPtr->flag("VinciaShower:"+parName);
  else if (settingsPtr->isFlag("Vincia:"+parName)) 
    par = settingsPtr->flag("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(name()+":setParm","ok: "+parName); }
  else {if (verbose >= 1) printErr(name()+":setParm",parName+": not found"); }
  return found;
}

bool Antenna::setParm(int& par, string parName) {
  bool found=true;
  if (settingsPtr->isMode(name()+":"+parName)) 
    par = settingsPtr->mode(name()+":"+parName);
  else if (settingsPtr->isMode("VinciaShower:"+parName)) 
    par = settingsPtr->mode("VinciaShower:"+parName);
  else if (settingsPtr->isMode("Vincia:"+parName)) 
    par = settingsPtr->mode("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(name()+":setParm","ok: "+parName);}
  else {if (verbose>= 1) printErr(name()+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(double& par, string parName) {
  bool found=true;
  if (settingsPtr->isParm(name()+":"+parName)) 
    par = settingsPtr->parm(name()+":"+parName);
  else if (settingsPtr->isParm("VinciaShower:"+parName)) 
    par = settingsPtr->parm("VinciaShower:"+parName);
  else if (settingsPtr->isParm("Vincia:"+parName)) 
    par = settingsPtr->parm("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(name()+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(name()+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(string& par, string parName) {
  bool found=true;
  if (settingsPtr->isWord(name()+":"+parName)) 
    par = settingsPtr->word(name()+":"+parName);
  else if (settingsPtr->isWord("VinciaShower:"+parName)) 
    par = settingsPtr->word("VinciaShower:"+parName);
  else if (settingsPtr->isWord("Vincia:"+parName)) 
    par = settingsPtr->word("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(name()+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(name()+":setParm",parName+": not found");}
  return found;
}

// Read in massless Laurent coefficients from Settings database

void Antenna::readC() {

  if (verbose >= 5)
    printOut(name()+":readC","reading antenna coefficients ");

  // First clear antenna function finite coefficients 
  clearC();
 
  //------------------------------------------------------------------
  // 1st part: read in non-singular antenna coefficients (massless)  
  for (int ic=-1;ic<=LaurentIndexMax;ic++) {
    string baseName("LaurentC("+num2str(ic,0)+",");
    for (int jc=-1;jc<=LaurentIndexMax;jc++) {
      string CName(baseName+num2str(jc,0)+")"); 
      if (settingsPtr->isParm(name()+":"+CName)) {	
	if (setC(ic,jc,settingsPtr->parm(name()+":"+CName))) {
	  int verb = (getC(ic,jc) == 0.0) ? 9 : 6; 
	  if (verbose >= verb)
	    printOut(name()+":readC",
		     CName+" = "+num2str(getC(ic,jc),6));
	}
      } 
    } 
  }
}
  
  //--------------------------------------------------------------------------

  // setC
  // Set or modify a non-singular antenna coefficient

  bool Antenna::setC(int ic,int jc, double cIn) {
    if (cIn == 0) {
      if (LaurentC.find(ic) != LaurentC.end()) {
	if (LaurentC[ic].find(jc) != LaurentC[ic].end()) LaurentC[ic].erase(jc);
      } 
    }
    else LaurentC[ic][jc]=cIn;
    return true;
  }
  
  //--------------------------------------------------------------------------

  // getC
  // return an antenna coefficient

  double Antenna::getC(int ic,int jc) {
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    it=LaurentC.find(ic);
    if (it==LaurentC.end()) return 0.0;
    jt=LaurentC[ic].find(jc);
    if (jt==LaurentC[ic].end()) return 0.0;
    return LaurentC[ic][jc];
  }
  
  void Antenna::clearC() { 
    LaurentC.clear(); 
  }

  //--------------------------------------------------------------------------

  // aBar
  // Function to get value of dimensionful antenna function
  // Defined as function of dimensionful (squared) invariants (linear) masses

  double Antenna::aBar(double sij, double sjk, double s,  
		       double massi, double massj, double massk, 
		       bool forcePositive) {
    if ( s == 0. ) return 0.0;
    double yij = sij/s;
    double yjk = sjk/s;
    if ( !useMassTerms || (massi <= 0. && massj <= 0. && massk <= 0.) ) 
      return aBarY(yij,yjk,0.,0.,0.,forcePositive)/s;
    else {
      double mIK = sqrt(s);
      double mui = massi/mIK;
      double muj = massj/mIK;
      double muk = massk/mIK;
      return aBarY(yij,yjk,mui,muj,muk,forcePositive)/s;
    }
  }    

  //--------------------------------------------------------------------------

  // aBarY
  // Function to get value of dimensionless antenna function
  // Defined as function of dimensionless squared invariants, yij = sij/s
  // and linear masses mui = mass(i)/s

  double Antenna::aBarY(double yij, double yjk, 
			double mui, double muj, double muk, 
			bool forcePositive) {

    // First check if switched off
    if ( !isOn() ) return 0.0;

    // Check masses
    bool massive = ( mui > 0. || muj > 0. || muk > 0. );
    if (massive && !useMassTerms) {
      mui = 0.;
      muj = 0.;
      muk = 0.;
      massive = false;
    }
    
    // Calculate the missing invariant
    double yik = 1. - yij - yjk;
    if (massive) yik += - pow2( mui ) - pow2( muj ) - pow2( muk );

    // Always check massless phase space boundary
    if (yik < 0. || yij < 0. || yjk < 0.) {
      return 0.;
    }

    // Perform more elaborate antenna checks if high verbose level
    if (verbose >= 3) {

      // complain if yik is below a value approaching machine precision
      if (yik - (massive ? 2*mui*muk : 0.0) < -TINY) {
	if (verbose>=3) {
	  printOut("VinciaAntenna::aBar", "yik < 2*mu1*mu3 occurred");
	  cout << "   yij = " << yij ;
	  cout << " yjk = " << yjk ;
        cout << " yik = " << yik << "\n";
        cout << "   mu1 = " << mui ;
        cout << " mu2 = " << muj ;
        cout << " mu3 = " << muk <<endl;      
	}
	return 0.0;
      } 
      // Tolerate values slightly below yikmin (due to rounding), 
      // but set them = 2*mui*muk
      else if (yik < 2*mui*muk) {
	yik = 2*mui*muk;
      }
      
      // MR 21.12.2010: Use Gram determinant to check whether we are too
      // close to the boundary of the phase space.
      // Check for gDet / Ecm^6 < TINY^3 because the Gram determinant has 
      // mass dimension 6
      // have to check yijmin <= yij <= yijmax first because
      // for yij > yijmax, a region with positive gram
      // determinant opens up
      double yijmin = 2*mui*muj;
      double yijmax = pow2( 1-muk ) - pow2(mui) - pow2(muj);
      if ( (yij - yijmin < 0.5*TINY ) || (yijmax - yij < 0.5*TINY) ) {
	if (verbose >= 3) {
	  printOut(name()+":aBar",
		   "yij = " + num2str(yij, 4) + " outside physical phase "
		   + "space or too close to boundaries yijmin = " 
		   + num2str(yijmin, 4) + ", yijmax = " 
		   + num2str(yijmax, 4) + "}, return 0.");
	}
	return 0.0;
      }
      double gdetdimless = gDet(yij, yjk, yik, mui, muj, muk);
      if (gdetdimless< pow3(TINY)) {
	if (verbose >= 3) {
	  printOut(name()+":aBar", 
		   "Gram determinant " + num2str(gdetdimless, 15) 
		   + " < " + num2str(pow3(TINY), 15)  + ", return 0\n"
		   + "\tyij = " + num2str(yij, 15) + ",\n"
		   + "\tyjk = " + num2str(yjk, 15) + ",\n"
		   + "\tyik = " + num2str(yik, 15) + ",\n"
		   + "\tmu1 = " + num2str(mui, 15) + ",\n"
		   + "\tmu2 = " + num2str(muj, 15) + ",\n"
		   + "\tmu3 = " + num2str(muk, 15)
		   );
	}
	return 0.0;
      }
    }

    // Compute antenna function value, singular and non-singular parts
    double val = antennaFunction(yij,yjk,yik,mui,muj,muk);

    // Check positivity requirement and return
    if (val >= 0 || !forcePositive) return val;
    else return 0.0;

  }
 
  //--------------------------------------------------------------------------

  // aEikonalY
  // Hardcoded function that gives the dimensionless eikonal.
  // Includes a mass correction for heavy quarks

  double Antenna::aEikonalY(double yij, double yjk, double yik,
			   double mui, double muk) {

    // Check denominator > 0
    double yT = yij*yjk;
    if (yT <= 0.) return 0.;

    // Compute eikonal, with mass correction
    double cEik    = getC(-1,-1);
    double eikonal = cEik*yik/(yij*yjk);
    if (useMassTerms && (mui > 0. || muk > 0.) )
      eikonal += - cEik*pow2(mui)/pow2(yij) - cEik*pow2(muk)/pow2(yjk);    

    // Return
    return eikonal;
    
  }

  //--------------------------------------------------------------------------

  // aTaylorY
  // Hardcoded function to sum up non-singular massless LaurentC coefficients

  double Antenna::aTaylorY(double yij, double yjk) {
    double finite = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // Only add finite terms (coeffs with both indices >= 0)
          if (ic >= 0 && jc >= 0) {
            double c=jt->second;
            // average c with charge-conjugate coefficient
            if (ic != jc) {
              c = 0.5*( c + getC(jc,ic) );
            }  
            finite += c * pow(yij,ic) * pow(yjk,jc);
          }
        }      
      }
    }
    return finite;    
  }

  //--------------------------------------------------------------------------

  // antennaFunction
  // Function encoding the (dimensionless) antenna function structure
  // normalized by the mass of the parent antenna.
  // To be modified in derived classes.

  double Antenna::antennaFunction(double yij, double yjk, double yik,
			 double mui, double muj, double muk) {

    // Check gluon massless
    if (muj > 0.) printOut(name()+"aSingY","using default eikonal but mass(j) > 0");

    // Return
    return aEikonalY(yij, yjk, yik, mui, muk);
    
  }

  //--------------------------------------------------------------------------

  // check
  // MR 21.12.2010: Changed the check function to rely on IsPositive
  // and IsFreeOfDeadZones. This loses a factor of 2 in efficiency
  // (the same phase space point is checked twice, once by IsPositive,
  // once by IsFreeOfDeadZones), but it makes the function easier to read.

  bool Antenna::check(int nPointsCheck, double mu1, double mu2, double mu3, 
		      bool dorepair) {
    bool isPositive = 
          this->IsPositive(nPointsCheck, mu1, mu2, mu3, dorepair);
    bool isFreeOfDeadZones = 
          this->IsFreeOfDeadZones(nPointsCheck, mu2, mu2, mu3, dorepair);
    return (isPositive && isFreeOfDeadZones);
  }
  
  //--------------------------------------------------------------------------

  // isPositive
  // MR 21.12.2010: IsPositive checks whether an antenna is positive
  // given a set of masses. It has an optional argument dorepair (default 
  // false) which governs whether the finite coefficients are changed 
  // to make the antenna positive.

  bool Antenna::IsPositive(int nPointsCheck, double mu1, double mu2, double mu3, 
			   bool dorepair) {
  // initialize the flags
  bool isPos = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  // calculate the phase space boundaries
  double yijmin = 2*mu1*mu2;
  double yijmax = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double yijStepWidth = (yijmax - yijmin)/numberofsteps;
  // i = 0 and i <= numberofsteps are on the boundary of the phase
  // space, omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double yij = yijmin + i*yijStepWidth;
    double yjkmin = y23tildeMin(yij, mu1, mu2, mu3);
    double yjkmax = y23tildeMax(yij, mu1, mu2, mu3);
    double yjkStepWidth = (yjkmax - yjkmin)/numberofsteps;
    // j=0 and j=numberofsteps are on the boundary of the phase space,
    // omit them
    for (int jc=1; jc<=numberofsteps-1; jc++) {
      double yjk = yjkmin + jc*yjkStepWidth;
      // Evaluate aBar with and without finite terms, enable
      // aBar to return negative values
      double aFull   = aBar(yij,yjk, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = aBar(yij,yjk, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      if ( aFull <= - TINY ) {
	isPos = false;
	if (dorepair) {
	  // if the singular part of the antenna is negative 
	  // and the function is supposed to repair the 
	  // antenna, there is a problem
	  if (aSing <= -TINY) {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name()+"aBar:IsPositive",
		       string("Singular part of antenna ")
		       + " negative.\n"
		       + "\tyij = " + num2str(yij, 4) +"\n"
		       + "\tyjk = " + num2str(yij, 4) +"\n"
		       + "\tmu1 = " + num2str(mu1, 4) +"\n"
		       + "\tmu2 = " + num2str(mu2, 4) +"\n"
		       + "\tmu3 = " + num2str(mu3, 4) +"\n"
		       + "\taFull = " + num2str(aFull, 4) +"\n"
		       + "\taSing = " + num2str(aSing, 4) +"\n"
		       );
	    }
	    verbose = verboseSav;
	    return false;
	  }
	  else {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 5) {
	      printOut(name()+":IsPositive",
		       "rescaling finite terms to make "
		       + string("antenna positive")
		       );
	    }
	    bool rescaleSucessful =
	      this->rescaleFiniteTerms(0.9);
	    if (rescaleSucessful) {
	      // restart the check with the changed 
	      // finite terms
	      if (verboseSav >= 3) {
		printOut(name()+":IsPositive",
			 "new coefficients are:");
		this->listCoefficients();
	      }
	      verbose = verboseSav;
	      return this->IsPositive(nPointsCheck, mu1, 
				      mu2, mu3, true);
	    }
	    else {
	      // if it has not been possible to rescale
	      // the finite terms, complain and return
	      // false
	      // verbose has been set to -1, 
	      // use verboseSav
	      if (verboseSav >= 1) {
		printOut(name()+"aBar:IsPositive",
			 "Could not rescale finite "
			 + string("terms.\n") );
	      }
	      verbose = verboseSav;
	      return false;
	    }
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsPositive::negative value" << endl;
	    cout << "Antenna name = " << humanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "yij = " << yij << endl;
	    cout << "yjk = " << yjk << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	  }
	}
      }
    } // end of for (int j=0; j<=numberofsteps...
  }
  // reset the verbosity
  verbose = verboseSav;
  return isPos;
}

  //--------------------------------------------------------------------------

  // isFreeOfDeadZones
  // MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has dead
  // zones given a set of masses. It has an optional argument dorepair 
  // (default false) which governs whether the finite coefficients are 
  // changed to make the dead zones go away.
  
  bool Antenna::IsFreeOfDeadZones( int nPointsCheck, double mu1, 
				   double mu2, double mu3, bool dorepair ) {
  // initialize the flags
  bool isFreeOfDeadZones = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  
  // calculate the phase space boundaries
  double yijmin = 2*mu1*mu2;
  double yijmax = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double yijStepWidth = (yijmax - yijmin)/numberofsteps;
  // i=0 and i=numberofsteps are the boundary of the phase space,
  // omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double yij = yijmin + i*yijStepWidth;
    // make sure yij is positive
    yij = max(yij,0.);
    double yjkmin = y23tildeMin(yij, mu1, mu2, mu3);
    double yjkmax = y23tildeMax(yij, mu1, mu2, mu3);
    double yjkStepWidth = (yjkmax - yjkmin)/numberofsteps;
    // j=0 and j=numberofsteps are the boundary of the phase space,
    // omit them
    for (int jc=1; jc<=numberofsteps-1; jc++) {
      double yjk = yjkmin + jc*yjkStepWidth;
      // make sure yjk is positive
      yjk = max(yjk, 0.);
      // Evaluate aBar with and without finite terms, enable
      // aBar to return negative values
      double aFull   = aBar(yij,yjk, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = aBar(yij,yjk, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      // only do something if aFull< deadZoneAvoidance*aSing
      if ( aFull < deadZoneAvoidance*aSing ) {
	isFreeOfDeadZones = false;
	if (dorepair) {
	  // rescale
	  if ( this->rescaleFiniteTerms(0.9) ) {
	    // restart the check with the changed 
	    // finite terms
	    verbose = verboseSav;
	    return this->IsFreeOfDeadZones(nPointsCheck, 
					   mu1, mu2, mu3, true);
	  }
	  else {
	    // if it has not been possible to rescale
	    // the finite terms, complain and return
	    // false
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name()+"aBar:IsFreeOfDeadZones",
		       "Could not rescale finite terms.\n");
	    }
	    verbose = verboseSav;
	    return false;
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsFreeOfDeadZones::dead zone" << endl;
	    cout << "Antenna name = " << humanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "yij = " << yij << endl;
	    cout << "yjk = " << yjk << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	    cout << "dead Zone avoidance factor = " 
		 << deadZoneAvoidance << endl;
	  }
	}
      }
    }
  }
  // reset the verbosity
  verbose = verboseSav;
  return isFreeOfDeadZones;
}

  //--------------------------------------------------------------------------

  // rescaleFiniteTerms

  bool Antenna::rescaleFiniteTerms(double rescalingFactor) {

    // list the old parameter values
    if (verbose >= 5) {
      printOut(name()+":rescaleFiniteTerms", 
	       string("reducing magnitude of finite terms.\n")
	       + "old parameter values:\n"
	       );
      listCoefficients();
    }
    // scale C(i,j) terms
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    for (it=LaurentC.begin();it!=LaurentC.end();++it) {
      int ic=it->first;
      for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
	int jc=jt->first;  
	if ( ic >= 0 && jc >= 0) {
	  double oldcoeff = jt->second;
	  double newabs = rescalingFactor*abs(oldcoeff);
	  // round the new absolute value down to the closest
	  // multiple of abs(1-rescalingFactor)
	  double minchange = abs(1-rescalingFactor);
	  newabs = floor( newabs/minchange )*minchange;
	  setC(ic, jc,oldcoeff*newabs/abs(oldcoeff) );
	}
      }
    } // end of loop through C(i,j) terms
    return true;
  }

  //--------------------------------------------------------------------------

  // listCoefficients
  // MR 16.7.2010: for debugging purposes, have the antenna list
  // its coefficients on an output stream
  
  void Antenna::listCoefficients(ostream& os) {
  os << setprecision(3);
  os << fixed;
  os << "antenna " << name() << "\n";
  
  os << "use finite terms: ";
  if (useFiniteTerms)  os << "True" << endl;
  else os << "False" << endl;
  
  map<int,map<int,double> >::iterator it;
  map<int,double>::iterator jt;
  // C(i,j) terms, first the fixed ones, then the finite ones
  for (it=LaurentC.begin();it!=LaurentC.end();++it) {
    int i=it->first;
    for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
      int jc=jt->first;  
      double c=jt->second;
      if (i < 0 || jc < 0) {
        os << "LaurentC(" << i << "," << jc << ") = " << c << "\n";
      }
    }      
  } 
  for (it=LaurentC.begin();it!=LaurentC.end();++it) {
    int i=it->first;
    for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
      int jc=jt->first;  
      double c=jt->second;
      if (i >= 0 && jc >= 0) {
        os << "LaurentC(" << i << "," << jc << ") = " << c << "\n";
      }
    }      
  }
}

  //--------------------------------------------------------------------------

  // gettestMasses
  // test masses for Antenna base are {0,0,0}
  vector< vector<double> > Antenna::gettestMasses() {
    vector< vector<double> > retVec;
    retVec.push_back( vector<double>(3, 0.0) );
    return retVec;
  }

  //==========================================================================
  // BEGIN FUNCTION DEFINITIONS FOR DERIVED CLASSES
  //==========================================================================

  //--------------------------------------------------------------------------

  // QQEmit
    
  void QQEmit::readC() {
    Antenna::readC();
    
    // read also useMassTerms
    if ( settingsPtr->isFlag(name() + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name() + ":UseMassTerms");
    }
    
    // use getParms
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name()+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu3ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yijord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int yjkord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setM(mu1ord,mu3ord,yijord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu1ord, mu3ord, yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+" = "+num2str(getM(mu1ord, mu3ord, yijord, yjkord),6)
          );
      }
    }
    const string 
      denominatorMassPrefactorName("DenominatorMassPrefactor");
    if (settingsPtr->isParm(name() + ":" + denominatorMassPrefactorName)){
      if (
        setDenominatorMassPrefactor(
          settingsPtr->parm(name()+":"+denominatorMassPrefactorName))){
        int verb = (denominatorMassPrefactor == 0.0) ? 8 : 6;
        if (verbose >= verb) {
          printOut(humanName()+":readC",
            denominatorMassPrefactorName + "=" 
            + num2str(getDenominatorMassPrefactor(), 6) );
        }
      }
    }
  }
  
  void QQEmit::clearC(){
    // Clear Laurent series coeffiecents
    Antenna::clearC();
    LaurentM.clear();
    setDenominatorMassPrefactor(0.0);
  }  
  
  void QQEmit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    os << "denominator mass prefactor: " << denominatorMassPrefactor
      << endl;
    for (map<FourInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      FourInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mIt->second << endl;
    }
  }
  
  bool QQEmit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    for (map< FourInt, double>::iterator pt=LaurentM.begin();
	 pt!=LaurentM.end();++pt) {
      int mu1ord = pt->first.int1;
      int mu3ord = pt->first.int2;
      int yijord = pt->first.int3;
      int yjkord = pt->first.int4;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setM(
	   mu1ord, mu3ord, yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    return baseOK;
  }
  
  double QQEmit::getM(
    int mu1ord, int mu3ord, int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentM.find(FourInt(mu1ord, mu3ord, yijord, yjkord));
    if (pos == LaurentM.end()) return 0.0;    
    else return pos->second;
  }
  
  bool QQEmit::setM(
    int mu1ord, int mu3ord, int yijord, int yjkord, double mIn
    ){
    FourInt index(mu1ord, mu3ord, yijord, yjkord);
    if (mIn == 0.0) LaurentM.erase(index);
    else LaurentM[index] = mIn;
    return true;
  }
  
  double QQEmit::getDenominatorMassPrefactor() {
    return denominatorMassPrefactor;
  }
  
   bool QQEmit::setDenominatorMassPrefactor(double pfIn) {
     denominatorMassPrefactor=pfIn;
     return true;
   }
  
  vector< vector<double> > QQEmit::gettestMasses() {
    // masses for a03 (m1,0,m2)
      const double a03_masses_input[][2] = 
      {	
        {0,0}, {0.01,0}, {0.05,0}, {0.3,0}, {0.5,0}, {0.9,0}, {0.95,0},
        {0.99,0}, {1,0},
        {0.01,0.01}, {0.05,0.05}, {0.3,0.3}, {0.45,0.45}, {0.49,0.49},
        {0.01,0.3}, {0.05,0.3}, {0.1,0.3},
        {0.01,0.95}, {0.01,0.99}, {0.05,0.9}, {0.05,0.95}, 
        {0.1,0.8}, {0.1,0.9}, {0.3,0.6}, {0.3,0.7}
      };
    vector< vector<double> > a03_masses;
    for (size_t i=0; i<sizeof(a03_masses_input)/(2*sizeof(double)); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(a03_masses_input[i][0]);
      prov_masses.push_back(0);
      prov_masses.push_back(a03_masses_input[i][1]);
      a03_masses.push_back( prov_masses );
    }
    return a03_masses;
  }

  double QQEmit::antennaFunction(double yij, double yjk, double yik, 
				 double mu1, double, double mu3){

    // First contribution: soft eikonal
    double eikonal     = aEikonalY(yij,yjk,yik,mu1,mu3);

    // Denominator for collinear and finite terms
    double denominator = yij+yik+yjk + denominatorMassPrefactor*mu1*mu3;

    // Non-eikonal collinear singularities 
    double collinear   = 0.0;
    if (useCollinearTerms) collinear = yij/yjk + yjk/yij;

    // Non-singular terms
    double finite      = 0.0;
    if (useFiniteTerms) {

      // Massless terms: 
      // Start from inherited Taylor part (finite LaurentC coefficients)
      finite = aTaylorY(yij,yjk);
      
      // Add antenna-specific massive terms
      if (useMassTerms && (mu1 > 0. || mu3 > 0.) ) { 
	for (map< FourInt, double>::iterator mIt=LaurentM.begin();
	     mIt!=LaurentM.end();++mIt) {
	  int mu1ord = mIt->first.int1;
	  int mu3ord = mIt->first.int2;
	  int yijord = mIt->first.int3;
	  int yjkord = mIt->first.int4;
	  double presentcoeff = mIt->second;
	  // ignore negative yij or yjk or exponents silently, 
	  // as well as nonpositive mu1ord and mu3ord 
	  if (min(yijord, yjkord) >= 0 && mu1ord >= 0 && mu3ord >= 0
	      && max(mu1ord, mu3ord) >= 1) {
	    finite += presentcoeff * pow(yij, yijord)*pow(yjk, yjkord) 
	      * pow(mu1, mu1ord) * pow(mu3, mu3ord);
	    // enforce charge conjugation symmetry by adding the 
	    // mu1ord <-> mu3ord coefficient 
	    // for non-symmetric coefficients
	    if (mu1ord != mu3ord) {
	      finite += presentcoeff*pow(yij, yijord)*pow(yjk, yjkord)
		*pow(mu1, mu3ord)*pow(mu3, mu1ord);
	    }
	    // analogous for yijord <-> yjkord
	    if (yijord != yjkord) {
              finite += presentcoeff*pow(yij, yjkord)*pow(yjk, yijord)
		*pow(mu1, mu1ord)*pow(mu3, mu3ord);
	    }
	    // and finally the combined switch
	    if (mu1ord != mu3ord && yijord != yjkord) {
	      finite += presentcoeff*pow(yij, yjkord)*pow(yjk, yijord)
		*pow(mu1, mu3ord)*pow(mu3, mu1ord);
	    }
	  }
	}
     }  // end of if (useMassTerms)
    } // end of if (useFiniteTerms)
      
    // Add collinear and finite terms to eikonal and return
    return eikonal + 1/denominator*(collinear + finite);

  }
  
  //--------------------------------------------------------------------------

  // QGEmit
    
  void QGEmit::readC() {
    Antenna::readC();
    
    // use getParms
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name()+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int muord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int yijord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yjkord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      setM(muord,yijord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(muord, yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+" = "+num2str(getM(muord, yijord, yjkord),6)
          );
      }
    }
    
    // analogous for the mtilde coefficients
    map<string,Parm> mtildecoeffs = 
      settingsPtr->getParmMap(name()+":LaurentMtilde(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mtildecoeffs.begin();
      pt!=mtildecoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int muord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int yijord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yikord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int yjkord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setMtilde(muord,yijord,yikord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getMtilde(muord,yijord,yikord,yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+" = "+num2str(getMtilde(muord,yijord,yikord,yjkord),6)
          );
      }
    }
  }
  
  void QGEmit::clearC(){
    // Clear Laurent series coeffiecents
    Antenna::clearC();
    LaurentM.clear();
    LaurentMtilde.clear();
  }  
  
  void QGEmit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    os << "collinear gluon splitting: " << octetPartitioning
      << endl;
    for (map<ThreeInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      ThreeInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << "): " 
        << mIt->second << endl;
    }
    for (map<FourInt, double>::iterator mtildeIt = LaurentMtilde.begin(); 
      mtildeIt != LaurentMtilde.end(); ++mtildeIt
      ) {
      FourInt indx = mtildeIt->first;
      os << "LaurentMtilde(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mtildeIt->second << endl;
    }
  }
  
  bool QGEmit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    for (map< ThreeInt, double>::iterator pt=LaurentM.begin();
	 pt!=LaurentM.end();++pt) {
      int muord = pt->first.int1;
      int yijord = pt->first.int2;
      int yjkord = pt->first.int3;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setM(muord,yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    for (map< FourInt, double>::iterator pt=LaurentMtilde.begin();
	 pt!=LaurentMtilde.end();++pt) {
      int muord = pt->first.int1;
      int yijord = pt->first.int2;
      int yikord = pt->first.int3;
      int yjkord = pt->first.int4;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
        setMtilde(
		  muord,yijord, yikord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    return baseOK;
  }
  
  double QGEmit::getM(
    int muord, int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<ThreeInt, double>::iterator pos = 
      LaurentM.find(ThreeInt(muord, yijord, yjkord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool QGEmit::setM(
    int muord, int yijord, int yjkord, double mIn
    ){
    ThreeInt index(muord, yijord, yjkord);
    if (mIn == 0.0) {
      LaurentM.erase(index);
    }
    else {
      LaurentM[index] = mIn;
    }
    return true;
  }
  
  double QGEmit::getMtilde(
    int muord, int yijord, int yikord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentMtilde.find(FourInt(muord, yijord, yikord, yjkord));
    if (pos == LaurentMtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool QGEmit::setMtilde(
    int muord, int yijord, int yikord, int yjkord, double mIn
    ){
    FourInt index(muord, yijord, yikord, yjkord);
    if (mIn == 0.0) {
      LaurentMtilde.erase(index);
    }
    else {
      LaurentMtilde[index] = mIn;
    }
    return true;
  }
  
  vector< vector<double> > QGEmit::gettestMasses() {
     // masses for d03 (m,0,0)
    const double d03_masses_input[] = 
      {0, 0.01, 0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95, 0.99, 1};
    vector< vector<double> > d03_masses;
    for (size_t i=0; i<sizeof(d03_masses_input)/sizeof(double); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(d03_masses_input[i]);
      prov_masses.push_back(0);
      prov_masses.push_back(0);
      d03_masses.push_back( prov_masses );
    }
    return d03_masses;
  }
  
  double QGEmit::antennaFunction(double yij, double yjk, double yik, 
				 double mu1, double, double) {

    // First contribution: soft eikonal
    double eikonal     = aEikonalY(yij,yjk,yik,mu1,0.);

    // Denominator for collinear and finite terms
    double denominator = yij + yik + yjk;

    // Non-eikonal collinear singularities
    double collinear   = 0.0;
    if ( useCollinearTerms ) collinear = 
                                yjk/yij + yij/yjk*(1-yij/denominator)
                                  + octetPartitioning*(yij-yik)/yjk;

    // Non-singular terms
    double finite      = 0.0;
    if ( useFiniteTerms ) {

      // Massless terms: 
      // Start from inherited Taylor part (finite LaurentC coefficients)
      finite = aTaylorY(yij,yjk);
      
      // Add antenna-specific massive terms
      if ( useMassTerms && mu1 > 0. ) { 
        for (map< ThreeInt, double>::iterator mIt=LaurentM.begin();
             mIt!=LaurentM.end();++mIt) {
          int muord = mIt->first.int1;
          int yijord = mIt->first.int2;
          int yjkord = mIt->first.int3;
          double presentcoeff = mIt->second;
          // ignore yij^(-n) silently except muq^(2+x)/yij
          // ignore yjk^(-1) silently
          // require positive mass order
          if (
	      yjkord >= 0 && muord >= 1 &&
	      (yijord >= 0 || (yijord == -1 && muord >= 2) )
	      ) {
            finite += 
              presentcoeff * pow(yij, yijord)*pow(yjk, yjkord) 
              * pow(mu1, muord);
          }
        }
        // add Mtilde terms
        for (
          map< FourInt,double>::iterator mtildeIt=LaurentMtilde.begin();
	  mtildeIt!=LaurentMtilde.end();++mtildeIt
          ) {
          int muord = mtildeIt->first.int1;
          int yijord = mtildeIt->first.int2;
          int yikord = mtildeIt->first.int3;
          int yjkord = mtildeIt->first.int4;
          double presentcoeff = mtildeIt->second;
          // only mu^(2+x)/yij/yik is allowed, other exponents of 
          // yij and yik are ignored silently
          if (
	      yjkord >= 0 && yijord == -1 && yikord ==-1 && muord >= 2 
            ) {
            finite += 
              presentcoeff * pow(yij, yijord) * pow(yik, yikord) 
              * pow(yjk, yjkord) * pow(mu1, muord);
          }
        }
      } // end of if (useMassTerms)
      
    } // end of if (useFiniteTerms)

    // Return sum of terms
    return eikonal + 1/denominator*(collinear + 1/denominator*finite);
  }  
  
  //--------------------------------------------------------------------------

  // GGEmit    
  
  double GGEmit::antennaFunction(double yij, double yjk, double yik, 
				  double, double, double) {

    // First contribution: soft eikonal
    double eikonal = 2*yik/(yij*yjk);

    double collinear = 0.0;
    if (useCollinearTerms) {
      collinear = yij/yjk + yjk/yij - pow2(yij)/yjk  - pow2(yjk)/yij 
	+ octetPartitioning*( (yij-yik)/yjk + (yjk-yik)/yij);
    }

    // Non-singular terms
    double finite      = 0.0;
    if ( useFiniteTerms ) finite = aTaylorY(yij,yjk);

    // Return the sum of temrs
    return eikonal + collinear + finite;

  }
  
  //--------------------------------------------------------------------------

  // QGSplit    
    
  void QGSplit::readC() {
    Antenna::readC();
    // read also useMassTerms
    if ( settingsPtr->isFlag(name() + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name() + ":UseMassTerms");
    }
    
    // use getParms
    // have to enter the bracket to not catch LaurentMtilde
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name()+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yijord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int yjkord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setM(mu1ord,mu23ord,yijord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu1ord, mu23ord, yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+" = "+num2str(getM(mu1ord, mu23ord, yijord, yjkord),6)
          );
      }
    }
    // read Ctilde and Mtilde in analogy
    map<string,Parm> mtildecoeffs = 
      settingsPtr->getParmMap(name()+":LaurentMtilde(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mtildecoeffs.begin();
      pt!=mtildecoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yijord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int yjkord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setMtilde(mu1ord,mu23ord,yijord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getMtilde(mu1ord, mu23ord, yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+
            " = "+num2str(getMtilde(mu1ord, mu23ord, yijord, yjkord),6)
          );
      }
    }
    map<string,Parm> ctildecoeffs = 
      settingsPtr->getParmMap(name()+":LaurentCtilde");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=ctildecoeffs.begin();
      pt!=ctildecoeffs.end();++pt
      ) {
      // read in the two numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int yijord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int yjkord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      setCtilde(yijord, yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getCtilde(yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+
            " = "+num2str(getCtilde(yijord, yjkord),6)
          );
      }
    }
  }
  
  void QGSplit::clearC(){
    // Clear Laurent series coeffiecents
    Antenna::clearC();
    LaurentM.clear();
    LaurentMtilde.clear();
    LaurentCtilde.clear();
  }  
  
  void QGSplit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    for (map<FourInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      FourInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mIt->second << endl;
    }
    for (
      map<FourInt, double>::iterator mtildeIt = LaurentMtilde.begin(); 
      mtildeIt != LaurentMtilde.end(); ++mtildeIt
      ) {
      FourInt indx = mtildeIt->first;
      os << "LaurentMtilde(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mtildeIt->second << endl;
    }
    for (
      map<pair<int,int>, double>::iterator cIt = LaurentCtilde.begin(); 
      cIt != LaurentCtilde.end(); ++cIt
      ) {
      pair<int,int> indx = cIt->first;
      os << "LaurentCtilde(" << indx.first << ", " << indx.second 
        << "): " 
        << cIt->second << endl;
    }
  }
  
  bool QGSplit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    // LaurentM
    for (map< FourInt, double>::iterator pt=LaurentM.begin();
	 pt!=LaurentM.end();++pt) {
      int muqord = pt->first.int1;
      int muqpord = pt->first.int2;
      int yijord = pt->first.int3;
      int yjkord = pt->first.int4;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setM(muqord,muqpord, yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    // LaurentMtilde
    for (map< FourInt, double>::iterator pt=LaurentMtilde.begin();
	 pt!=LaurentMtilde.end();++pt) {
      int muqord = pt->first.int1;
      int muqpord = pt->first.int2;
      int yijord = pt->first.int3;
      int yjkord = pt->first.int4;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setMtilde(muqord,muqpord, yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    // LaurentCtilde
    for (
	 map< pair<int,int>, double>::iterator pt=LaurentCtilde.begin();
	 pt!=LaurentCtilde.end();++pt
	 ) {
      int yijord = pt->first.first;
      int yjkord = pt->first.second;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setCtilde(yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    return baseOK;
  }

  double QGSplit::getM(
    int muqord, int muqpord, int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentM.find(FourInt(muqord, muqpord, yijord, yjkord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool QGSplit::setM(
    int muqord, int muqpord, int yijord, int yjkord, double mIn
    ) {
    FourInt index(muqord, muqpord, yijord, yjkord);
    if (mIn == 0.0) {
      LaurentM.erase(index);
    }
    else {
      LaurentM[index] = mIn;
    }
    return true;
  }
  
  double QGSplit::getMtilde(
    int muqord, int muqpord, int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentMtilde.find(FourInt(muqord, muqpord, yijord, yjkord));
    if (pos == LaurentMtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool QGSplit::setMtilde(
    int muqord, int muqpord, int yijord, int yjkord, double mIn
    ) {
    FourInt index(muqord, muqpord, yijord, yjkord);
    if (mIn == 0.0) {
      LaurentMtilde.erase(index);
    }
    else {
      LaurentMtilde[index] = mIn;
    }
    return true;
  }
  
  double QGSplit::getCtilde(
    int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map< pair<int,int>, double>::iterator pos = 
      LaurentCtilde.find(make_pair(yijord, yjkord));
    if (pos == LaurentCtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool QGSplit::setCtilde(
    int yijord, int yjkord, double cIn
    ) {
    pair<int,int> index(yijord, yjkord);
    if (cIn == 0.0) {
      LaurentCtilde.erase(index);
    }
    else {
      LaurentCtilde[index] = cIn;
    }
    return true;
  }
  
  vector< vector<double> > QGSplit::gettestMasses() {
    const double e03_masses_input[][3] = 
    {	
      {0,0,0}, {0.01,0,0}, {0.05,0,0}, {0.3,0,0}, {0.5,0,0}, 
      {0.9,0,0}, {0.95,0,0}, {0.99,0,0}, {1,0,0},
      {0,0.01,0.01}, {0.01,0.01,0.01}, {0.05,0.01,0.01}, 
      {0.3,0.01,0.01}, {0.5,0.01,0.01}, {0.9,0.01,0.01}, 
      {0.95,0.01,0.01}, {0.98,0.01,0.01},
      {0,0.05,0.05}, {0.01,0.05,0.05}, {0.05,0.05,0.05}, 
      {0.3,0.05,0.05}, {0.5,0.05,0.05}, {0.85,0.05,0.05}, 
      {0.9,0.05,0.05}, 
      {0,0.3,0.3}, {0.01,0.3,0.3}, {0.05,0.3,0.3}, 
      {0.3,0.3,0.3}, {0.35,0.3,0.3}, {0.4,0.3,0.3}, 
      {0,0.45,0.45}, {0.01,0.45,0.45}, {0.05,0.45,0.45}, 
      {0.1,0.45,0.45}
    };
    vector< vector<double> > e03_masses;
    for (
      size_t i=0; i<sizeof(e03_masses_input)/(3*sizeof(double)); i++
      ) {
      vector<double> prov_masses;
      prov_masses.push_back(e03_masses_input[i][0]);
      prov_masses.push_back(e03_masses_input[i][1]);
      prov_masses.push_back(e03_masses_input[i][2]);
      e03_masses.push_back( prov_masses );
    }
    return e03_masses;
  }
  
  double QGSplit::antennaFunction(double yij, double yjk, double yik, 
				  double muq, double muqp1, double muqp2) {
    // silently average the secondary quark masses
    double muqp = 0.5*( muqp1 + muqp2 );
    double denominator = yjk + 2*pow2(muqp);
    // warning: if useMassTerms is false, aBarY is called with a sum 
    // of reduced yab and muc which do NOT add up to 1
    double yIK = yij + yik + yjk + 2*pow2(muqp);
    double fixed = 
      0.5/denominator*( (pow2(yik) + pow2(yij)) / pow2(yIK) + 2*pow2(muqp)/denominator);

    // Non-singular terms
    double finite      = 0.0;
    if ( useFiniteTerms ) {

      // Massless terms: 
      // Start from inherited Taylor part (finite LaurentC coefficients)
      finite = aTaylorY(yij,yjk);
      
      // Add antenna-specific massive terms
      if ( useMassTerms && (muq > 0.0 || muqp > 0.0) ) {
	// LaurentM
	for (
	     map< FourInt,double>::iterator mIt=LaurentM.begin();
	     mIt!=LaurentM.end();++mIt
	     ) {
	  int muqord = mIt->first.int1;
	  int muqpord = mIt->first.int2;
	  int yijord = mIt->first.int3;
	  int yjkord = mIt->first.int4;
	  double presentcoeff = mIt->second;
	  // only nonnegative powers of yij and yjk are allowed
	  // at least one positive power of muq or muqp is needed
	  if (
	      yijord >= 0 && yjkord >= 0 && max(muqord, muqpord) >= 1 
	      ) {
	    finite += 
	      yjk/denominator * presentcoeff * pow(yij, yijord) * pow(yjk, yjkord) 
	      * pow(muq, muqord) * pow(muqp, muqpord);
	  }
	}
	// LaurentMtilde
	for (
	     map< FourInt,double>::iterator mtildeIt=LaurentMtilde.begin();
	     mtildeIt!=LaurentMtilde.end();++mtildeIt
	     ) {
	  int muqord = mtildeIt->first.int1;
	  int muqpord = mtildeIt->first.int2;
	  int yijord = mtildeIt->first.int3;
	  int yjkord = mtildeIt->first.int4;
	  double presentcoeff = mtildeIt->second;
	  // only nonnegative powers of yij and yjk are allowed
	  // at least one positive power of muq or muqp is needed
	  if (
	      yijord >= 0 && yjkord >= 0 && max(muqord, muqpord) >= 1 
	      ) {
	    finite += 
	      pow2(muqp)/denominator * presentcoeff * pow(yij, yijord) * 
	      pow(yjk, yjkord) * pow(muq, muqord) * pow(muqp, muqpord);
	  }
	}
	// LaurentCtilde
	for (
	     map< pair<int,int>, double>::iterator cIt=LaurentCtilde.begin(); 
	     cIt != LaurentCtilde.end(); ++cIt
	     ) {
	  int yijord = cIt->first.first;
	  int yjkord = cIt->first.second;
	  double presentcoeff = cIt->second;
	  // only nonnegative powers are allowed
	  if (yijord >= 0 && yjkord >= 0) {
	    finite += 
	      pow2(muqp)/denominator * presentcoeff * pow(yij, yijord) * 
	      pow(yjk, yjkord);
	  }
	}
      }
    }

    // Return sum of terms
    return fixed + finite/pow2(yIK);

  }
    
  //--------------------------------------------------------------------------

  // GGSplit    
  
  void GGSplit::readC() {
    Antenna::readC();
    // read also useMassTerms
    if ( settingsPtr->isFlag(name() + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name() + ":UseMassTerms");
    }
    // read Ctilde and M
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name()+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu23ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int yijord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int yjkord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      setM(mu23ord,yijord,yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu23ord, yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+
            " = "+num2str(getM(mu23ord, yijord, yjkord),6)
          );
      }
    }
    map<string,Parm> ctildecoeffs = 
      settingsPtr->getParmMap(name()+":LaurentCtilde");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=ctildecoeffs.begin();
      pt!=ctildecoeffs.end();++pt
      ) {
      // read in the two numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int yijord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int yjkord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      setCtilde(yijord, yjkord, pt->second.valNow);
      // output
      int verb = 
        (abs(getCtilde(yijord, yjkord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName()+":readC",
          parname+
            " = "+num2str(getCtilde(yijord, yjkord),6)
          );
      }
    }
  }
  
  void GGSplit::clearC(){
    // Clear Laurent series coeffiecents
    Antenna::clearC();
    LaurentM.clear();
    LaurentCtilde.clear();
  }  
  
  void GGSplit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    for (
      map<ThreeInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      ThreeInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << "): " 
        << mIt->second << endl;
    }
    for (
      map<pair<int,int>, double>::iterator cIt = LaurentCtilde.begin(); 
      cIt != LaurentCtilde.end(); ++cIt
      ) {
      pair<int,int> indx = cIt->first;
      os << "LaurentCtilde(" << indx.first << ", " << indx.second 
        << "): " 
        << cIt->second << endl;
    }
  }
  
  bool GGSplit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    // LaurentM
    for (map< ThreeInt, double>::iterator pt=LaurentM.begin();
	 pt!=LaurentM.end();++pt) {
      int muqpord = pt->first.int1;
      int yijord = pt->first.int2;
      int yjkord = pt->first.int3;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setM(
	   muqpord, yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    // LaurentCtilde
    for (
	 map< pair<int,int>, double>::iterator pt=LaurentCtilde.begin();
	 pt!=LaurentCtilde.end();++pt
	 ) {
      int yijord = pt->first.first;
      int yjkord = pt->first.second;
      double oldcoeff = pt->second;
      double newabs = rescalingFactor*abs(oldcoeff);
      // round the new absolute value down to the closest
      // multiple of abs(1-rescalingFactor)
      double minchange = abs(1-rescalingFactor);
      newabs = floor( newabs/minchange )*minchange;
      setCtilde(yijord, yjkord, oldcoeff*newabs/abs(oldcoeff));
    }
    return baseOK;
  }
  
  double GGSplit::getM(
    int muqpord, int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map<ThreeInt, double>::iterator pos = 
      LaurentM.find(ThreeInt(muqpord, yijord, yjkord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool GGSplit::setM(
    int muqpord, int yijord, int yjkord, double mIn
    ) {
    ThreeInt index(muqpord, yijord, yjkord);
    if (mIn == 0.0) {
      LaurentM.erase(index);
    }
    else {
      LaurentM[index] = mIn;
    }
    return true;
  }
  
  double GGSplit::getCtilde(
    int yijord, int yjkord
    ) {
    // use find to avoid the creation of an entry
    map< pair<int,int>, double>::iterator pos = 
      LaurentCtilde.find(make_pair(yijord, yjkord));
    if (pos == LaurentCtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool GGSplit::setCtilde(
    int yijord, int yjkord, double cIn
    ) {
    pair<int,int> index(yijord, yjkord);
    if (cIn == 0.0) {
      LaurentCtilde.erase(index);
    }
    else {
      LaurentCtilde[index] = cIn;
    }
    return true;
  }
  
  vector< vector<double> > GGSplit::gettestMasses() {
    const double g03_masses_input[] =
      { 0., 0.01, 0.05, 0.1, 0.3, 0.45, 0.49, 0.5 };
    vector< vector<double> > g03_masses;
    for (size_t i=0; i<sizeof(g03_masses_input)/sizeof(double); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(0);
      prov_masses.push_back(g03_masses_input[i]);
      prov_masses.push_back(g03_masses_input[i]);
      g03_masses.push_back( prov_masses );
    }
    return g03_masses;
  }
  
  double GGSplit::antennaFunction(double yij, double yjk, double yik, 
				  double, double muqp1, double muqp2) {

    // silently average the secondary quark masses
    double muqp = 0.5*( muqp1 + muqp2 );
    double denominator = yjk + 2*pow2(muqp);
    double fixed = 
      0.5/denominator*( pow2(yik) + pow2(yij) + 2*pow2(muqp)/denominator);

    // Non-singular terms
    double finite      = 0.0;
    if ( useFiniteTerms ) {

      // Massless terms: 
      // Start from inherited Taylor part (finite LaurentC coefficients)
      finite = aTaylorY(yij,yjk);
      
      // Add antenna-specific massive terms
      if ( useMassTerms && muqp > 0.0 ) {

	// LaurentM
	for (
	     map< ThreeInt,double>::iterator mIt=LaurentM.begin();
	     mIt!=LaurentM.end();++mIt
	     ) {
	  int muqpord = mIt->first.int1;
	  int yijord = mIt->first.int2;
	  int yjkord = mIt->first.int3;
	  double presentcoeff = mIt->second;
	  // only nonnegative powers of yij and yjk are allowed
	  // at least one positive power of muq is needed
	  if (
	      yijord >= 0 && yjkord >= 0 && muqpord >= 1 
	      ) {
	    finite += 
	      pow2(muqp)/denominator * presentcoeff * pow(yij, yijord) 
	      * pow(yjk, yjkord) * pow(muqp, muqpord);
	  }
	}
	// LaurentCtilde
	for (
	     map< pair<int,int>, double>::iterator cIt=LaurentCtilde.begin();
	     cIt != LaurentCtilde.end(); ++cIt
	     ) {
	  int yijord = cIt->first.first;
	  int yjkord = cIt->first.second;
	  double presentcoeff = cIt->second;
	  // only nonnegative powers are allowed
	  if (yijord >= 0 && yjkord >= 0) {
	    finite += 
	      pow2(muqp)/denominator * presentcoeff * pow(yij, yijord) * 
	      pow(yjk, yjkord);
	  }
	}
      } // end if ( useMassTerms )
    } // end if ( useFiniteTerms ) 

    // Return sum of terms
    return fixed + finite;
  }

  
} // End Vincia Namespace
  
