// VinciaTools.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for VINCIA's Toolbox.

#include "VinciaTools.h"

// Declare the F77 3->2 clustering map for use below
extern"C" {
  void vc3to2_(int *MMAP, double PA[4], double PR[4], double PB[4],
	       double PAHAT[4], double PBHAT[4], int *IFAIL);
}

namespace Vincia {

//***********************************************************************

// VinClu: VINCIA Clustering 
// A set of 3->2 dipole-antenna mappings

// ******
// getClu : to recluster one parton inside a set of id+momenta 
// INPUT:
//   idIn : original vector of parton id (PDG) codes
//   pIn  : original vector of parton 4-momenta
//   ia, ir, ib : the 3 partons to be clustered, of which ir is to be unresolved
// OUTPUT:
//   idClu   : vector of clustered id (PDG) codes
//   pClu    : vector of clustered 4-momenta
//   antIndx : index of the corresponding antenna function in VINCIA
//   swapAB  : indicates whether the antenna function or its cc is to be evaluated
// Notes:
//   Mainly designed for the case ia = ir-1, ib = ir+1
//      (i.e., for ia,ir,ib ordered sequentially)
//   Can also be used for arbitrary ia, ir, ib, with the restriction that 
//      ia < ir < ib

  bool VinClu::getClu(vector<int> idIn, vector<Vec4> pIn,int ia, int ir, int ib, 
		      vector<int>& idClu, vector<Vec4>& pClu, 
		      int& antIndx, bool& swapAB) {
    
    if (verbose >= 8) printOut("VinClu::getClu","begin ---------");

    // Make sure we're initialized
    if (! isInit) init();

    // Start clustering
    idClu  = idIn;
    swapAB = false;
    int ida     = idIn[ia];
    int idr     = idIn[ir];
    int idb     = idIn[ib];
    int iaClu   = ia;
    int ibClu   = ib;
    if (idr == 21) {
      // To recluster gluon, just collapse parent ids
      idClu.erase(idClu.begin()+ir);
      if (ida == 21 && idb == 21) antIndx = iggEmit;
      else if (idb == 21) antIndx = iqgEmit;
      else if (ida == 21) {antIndx = iqgEmit; swapAB=true;}
      else antIndx = iqqEmit;
    } else if (idr < 0) {
      // To recluster antiquark, collapse with downstream quark 
      idClu[ibClu]=21;
      idClu.erase(idClu.begin()+ir);
      // gqbarq 
      if (ida == 21) {
	antIndx = iggSplit;
	swapAB  = false;
      }
      else {
	// qqbar'q' 
	antIndx = iqgSplit;
	swapAB  = false;
      }
    } else {
      // To recluster quark, collapse with upstream antiquark
      idClu[iaClu]=21;
      idClu.erase(idClu.begin()+ir);
      // qbarqg
      if (idb == 21) {
	antIndx = iggSplit; 
	swapAB=true;
      }
      // qbar'q'qbar
      else {
	antIndx = iqgSplit; 
	swapAB=true;
      }
    }
    // Now decrease ibClu by one, since ir has been erased.
    ibClu--;

    // Determine which kinematics map to use
    int iMap = settingsPtr->mode("Vincia:kineMapType");

    // Later: allow for antenna-specific kineMapTypes
    // ... 

    // Reconstruct clustered momenta
    bool clusterOK = true;

    // Dereference pointer to make sure it is ???
    //VinciaCommon vinComTmp = *vinComPtr;

    // Massless Case 
    if ( !vinComPtr->isMassive(ida) && !vinComPtr->isMassive(idb) ) {
      
      if (verbose >= 7) printOut("VinClu::getClu","doing massless reclustering");
      clusterOK = vc3to2(iMap,pIn,ia,ir,ib,pClu);
      
    } 

    // Massive Case 
    else {

      // Parent masses
      // If a heavy flavor goes through, compute dynamical mass
      // otherwise the parent mass is zero (gluons or light flavors)
      double mI=0.0;
      double mK=0.0;
      mI = (idIn[ia] == idClu[iaClu] && idIn[ia] != 21 && abs(idIn[ia]) >= 4) ? 
	pIn[ia].mCalc() : 0.0;
      mK = (idIn[ib] == idClu[ibClu] && idIn[ib] != 21 && abs(idIn[ia]) >= 4) ? 
	pIn[ib].mCalc() : 0.0;
      if (verbose >= 7) {
	printOut("VinClu::getClu","doing massive reclustering with mI = "
		 +num2str(mI)+" mK = "+num2str(mK));
      }
      // Only maps >= 3 implemented for massive case
      int kineMapTemp = max(3,iMap);
      // Do clustering      
      clusterOK = vc3to2massive(kineMapTemp,pIn,ia,ir,ib,mI,mK,pClu);
    }
    if (verbose >= 9 || (! clusterOK && verbose >= 3) ) {
      for (int i=0;i<int(pClu.size());i++) {
	cout << "   p" << i<<" "<<pClu[i];
      }
    }    

    // Return pass/fail
    return clusterOK;

  } // End getClu

  //******

  // vc3to2: main clustering map
  // Input:
  //   iMap    = map number ( 1 : ARIADNE, 2 : PS, 3 : Kosower)
  //   pIn     = Vec4 list (arbitrary length, but greater than 3)
  //   a,r,b   = indices of 3 particles in pIn to be clustered
  // Output:  
  //   pClu    = pIn but with the a and b momenta replaced by the clustered
  //             aHat and bHat and with r erased.
  // I.e.,
  //   pIn(...,a,...,r-1,r,r+1,...,b,...) 
  //     -> 
  //   pOut(...,aHat,...,r-1,r+1,...,bHat,...)
  //   with {a,r,b} -> {aHat,bHat} using kinematics map iMap
  
  bool VinClu::vc3to2(int iMap,vector<Vec4> pIn,int a, int r, int b, 
		      vector<Vec4>& pClu) {
    
    // Intialize and sanity check
    pClu=pIn;
    if (max(max(a,r),b) > int(pIn.size()) || min(min(a,r),b) < 0) {
      if (verbose >= 3) 
	printOut("VinClu::vc3to2","Error! Unable to cluster (a,r,b) = "+
		 num2str(a)+num2str(r)+num2str(b)+" p.size ="
		 +num2str(int(pIn.size()))); 
      return false;
    }    
    // Use fortran map for ARIADNE and PS maps
    // (maps -1 and -2 are special: force A or B to take all recoil)
    if (iMap == 1 || iMap == 2 || iMap == -1 || iMap == -2) {
      // Initialize variables for use in re-clusterings
      double pa[4],pr[4],pb[4],pahat[4],pbhat[4];
      int ifail=0;      

      // Do reclustering
      Vec4toD4(pIn[a],pa);
      Vec4toD4(pIn[r],pr);
      Vec4toD4(pIn[b],pb);
      /*
      cout<<" pa = "<<pa[0]<<" "<<pa[1]<<" "<<pa[2]<<" "<<pa[3]<<endl;
      cout<<" pr = "<<pr[0]<<" "<<pr[1]<<" "<<pr[2]<<" "<<pr[3]<<endl;
      cout<<" pb = "<<pb[0]<<" "<<pb[1]<<" "<<pb[2]<<" "<<pb[3]<<endl;
      */

      // Cluster
      vc3to2_(&iMap,pa,pr,pb,pahat,pbhat,&ifail);            
      // Check return code
      if (ifail != 0) {
	if (verbose >= 3) 
	  printOut("VinClu:vc3to2","vc3to2_ returned "+num2str(ifail,2));	
	return false;
      }
      
      // Store reclustered momenta
      pClu[a]=(Vec4(pahat[0],pahat[1],pahat[2],pahat[3]));
      pClu[b]=(Vec4(pbhat[0],pbhat[1],pbhat[2],pbhat[3]));
      pClu.erase(pClu.begin()+r);
      return true;
    } 
     
    // MR 11.7.2010: massless antenna mapping or dipole mapping
    else if (iMap == 3 || iMap == 4) {
      bool clusterOK = map3to2(iMap, pIn[a], pIn[r], pIn[b], pClu[a], pClu[b]);
      pClu.erase(pClu.begin()+r);      
      if (verbose >= 3 && !clusterOK ) 
	printOut("VinClu::vc3to2","map3to2 returned false");
      return clusterOK;
    }
    
    // Default to type 3 map if called with unknown type
    else {
      bool clusterOK = map3to2(3, pIn[a], pIn[r], pIn[b], pClu[a], pClu[b]);
      pClu.erase(pClu.begin()+r);      
      // Issue warning if asked to be verbose
      if (verbose >=3) 
	printOut("VinClu::vc3to2","called with unknown map type " + num2str(iMap));
      return clusterOK;
    }

    // Should not arrive here
    return false;
  }

  // MR 7.7.2010: define vc3to2massive in direct analogy to vc3to2 above

  // vc3to2massive: massive version of the main clustering map
  // Input:
  //   iMap    = map number ( 1 : ARIADNE, 2 : PS, 3 : Kosower)
  //   pIn     = Vec4 list (arbitrary length, but greater than 3)
  //   a,r,b   = indices of 3 particles in pIn to be clustered
  //   mI      = mass of first parent particle
  //   mK      = mass of second parent particle
  // Output:  
  //   pClu    = pIn but with the a and b momenta replaced by the clustered
  //             aHat and bHat and with r erased.
  // I.e.,
  //   pIn(...,a,...,r-1,r,r+1,...,b,...) 
  //     -> 
  //   pOut(...,aHat,...,r-1,r+1,...,bHat,...)
  //   with {a,r,b} -> {aHat,bHat} using kinematics map iMap
  
  bool VinClu::vc3to2massive(int iMap,vector<Vec4> pIn,int a, int r, int b, 
		      double mI, double mK, vector<Vec4>& pClu) {
    
    // if both parent masses are negligible and all the momenta are 
    // massless return vc3to2
    // measure off-shellness normalised to average energy of the 
    // partons to be clustered, avoids small denominator for 
    // collinear massless p_a, p_r, p_b 
    double eAv = 1.0/3.0*( pIn[a].e() + pIn[r].e() + pIn[b].e() );
    if (mI/eAv < TINY && mK/eAv < TINY && pIn[a].mCalc()/eAv < TINY 
          && pIn[r].mCalc()/eAv < TINY && pIn[b].mCalc()/eAv < TINY ) {
      return vc3to2(iMap, pIn, a, r, b, pClu);
    }
    
    // Intialize and sanity check
    pClu=pIn;
    if (max(max(a,r),b) > int(pIn.size()) || min(min(a,r),b) < 0) return false;
    bool clusterOK = false;

    // Massive case only implemented for maps >= 3
    int kineMapTemp = max(3,iMap);
    
    if (iMap == 3 || iMap == 4) {
      clusterOK = map3to2massive(kineMapTemp, pIn[a],pIn[r],pIn[b], mI, mK,
				 pClu[a],pClu[b]);
      if (verbose >=3 && !clusterOK) 
	printOut("VinClu::vc3to2masive","map3to2massive returned false");
    }
    else {
      if ( verbose >= 2 ) {
	printOut("VinClu::vc3to2massive", "unknown kinematics map type " + num2str(iMap));
      }
    }
    // erase the vector r
    pClu.erase(pClu.begin()+r);
    return clusterOK;
  }
  
  // ******
  
  // Kosower's 3to2 clustering map (as used by GGG in 4-parton functions)
  // MR 12.7.2010: replace with map3to2
  /*
    void VinClu::ks3to2(Vec4 pi, Vec4 pj, Vec4 pk, Vec4& pOutI, Vec4& pOutK) {
    map3to2(3, pi, pj, pk, pOutI, pOutK);
    // Compute invariants
    double sij  = m2(pi,pj);
    double sjk  = m2(pj,pk);
    double sik  = m2(pi,pk);
    double sijk = m2(pi,pj,pk);
    
    // Compute coefficients (choice of r fixed to the one used by GGG)
    double r    = sjk / (sij + sjk); 
    double rho  = sqrt(1.0+(4*r*(1.0-r)*sij*sjk)/sijk/sik);
    double x    = 0.5/(sij+sik)*((1+rho)*sijk-2*r*sjk);
    double z    = 0.5/(sjk+sik)*((1-rho)*sijk-2*r*sij);
    
    // Compute reclustered momenta
    pOutI       =     x*pi +     r*pj +     z*pk;
    pOutK       = (1-x)*pi + (1-r)*pj + (1-z)*pk;
    }
  */
  
  // MR 11.7.2010: write map3to2 which takes the type of the mapping
  // as its first argument
  // change vc3to2 and map3to2 to return booleans, return false if
  // the kinematicsType is unknown or something else wrong
  bool VinClu::map3to2(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
						Vec4& pOutI, Vec4& pOutK) {
    if (verbose >= 8) {
      printOut("VinClu::map3to2", "called with ");
      cout << "pi = " << pi;
      cout << "pj = " << pj;
      cout << "pk = " << pk;
    }
    
    // initialize the return value to true, it will be set to false
    // if the on-shell condition is violated
    bool clusterOK = true;
    
    // Compute invariants
    double sij  = 2*pi*pj;
    double sjk  = 2*pj*pk;
    double sik  = 2*pi*pk;
    // check whether the arguments need to be reversed for kineMapType==4
    if (kineMapType == 4 && ( ! (sij < sjk) ) ) {
      if (verbose >= 8) {
	printOut("VinClu::map3to2","choose parton i as the recoiler");
      }
      // call the function with reverse arguments, then return
      return map3to2(kineMapType, pk, pj, pi, pOutK, pOutI);
    }
    double sijk = sij + sik + sjk;
    
    double r = 0;
    if (kineMapType==3) {
      // Compute coefficients (choice of r fixed to the one used by GGG)
      r    = sjk / (sij + sjk); 
    }
    else if (kineMapType==4) {
      r = 1.0;
    }
    // if the kineMapType is not known, complain and return false
    else {
      if (verbose >=3) 
	printOut("VinClu::map3to2", "unknown kineMapType "
		 + num2str(kineMapType));
      pOutI = 0*pi;
      pOutK = 0*pi;
      return false;
    }
    
    double rho  = sqrt(1.0+(4*r*(1.0-r)*sij*sjk)/sijk/sik);
    double x    = 0.5/(sij+sik)*((1+rho)*(sij+sik)+(1+rho-2*r)*sjk);
    double z    = 0.5/(sjk+sik)*((1-rho)*sijk-2*r*sij);
    // Compute reclustered momenta
    pOutI       =     x*pi +     r*pj +     z*pk;
    pOutK       = (1-x)*pi + (1-r)*pj + (1-z)*pk;
    
    // MR 12.7.2010: check that I got massless momenta
    // MR 12.11.2010: normalise the off-shellness to sijk to get
    // a dimensionless quantitiy to compare with TINY
    if (pOutI.m2Calc()/sijk >= TINY || pOutK.m2Calc()/sijk >= TINY) {
      if (verbose>=3) 
	printOut("VinClu::map3to2","on-shell check failed. m2I/sIK ="
		 + num2str(pOutI.m2Calc()/sijk)+" m2K/sIK ="
		 + num2str(pOutK.m2Calc()/sijk)+" sIK = "+num2str(sijk));
      clusterOK = false;
    }    
    
    return clusterOK;
  }
  
  // MR 7.7.2010: massive generalization of 3->2 clustering
  bool VinClu::map3to2massive(int kineMapType, Vec4 pi, Vec4 pj, Vec4 pk,
			      double mI, double mK, Vec4& pOutI, 
			      Vec4& pOutK) {
    if (verbose>=8) {
      printOut("VinClu::map3to2massive", "called with ");
      cout << "pi = " << pi;
      cout << "pj = " << pj;
      cout << "pk = " << pk;
    }
    
    // if both parent masses are negligible and all the momenta are 
    // massless return vc3to2
    // measure off-shellness normalised to average energy of the 
    // partons to be clustered, avoids small denominator for 
    // collinear massless p_a, p_r, p_b 
    double eAv = 1.0/3.0*( pi.e() + pj.e() + pk.e() );
    if (mI/eAv < TINY && mK/eAv < TINY && pi.mCalc()/eAv < TINY 
          && pj.mCalc()/eAv < TINY && pk.mCalc()/eAv < TINY ) {
      return map3to2(kineMapType, pi, pj, pk, pOutI, pOutK);
    }
    
    // initialize the return value to true, it will be set false if
    // the on-shell check of the reclustered momenta fails
    bool clusterOK = true;
    
    // Compute invariants
    double mi = pi.mCalc();
    double mj = pj.mCalc();
    double mk = pk.mCalc();
    double sij  = 2*pi*pj;
    double sjk  = 2*pj*pk;
    double sik  = 2*pi*pk;
    // check whether the arguments need to be reversed for kineMapType==4
    if (kineMapType == 4 && (! (sij < sjk) ) ) {
      // call the function with reverse arguments, then return
      return map3to2massive(kineMapType, pk, pj, pi, 
			    mK, mI, pOutK, pOutI);
    }
    double Ecmsqua = sij + sjk + sik + pi.m2Calc() + pj.m2Calc() + pk.m2Calc();
    double Ecm = sqrt(Ecmsqua);
    
    // rewrite the determination in terms of dimensionless variables
    
    double mui = mi/Ecm;
    double muj = mj/Ecm;
    double muk = mk/Ecm;
    double yij = sij/Ecmsqua;
    double yjk = sjk/Ecmsqua;
    double yik = sik/Ecmsqua;
    
    double yijmin = 2*mui*muj;
    double yjkmin = 2*muj*muk;
    double yikmin = 2*mui*muk;
    
    double muI = mI/Ecm;
    double muK = mK/Ecm;
    double yIK = 1. - pow2(muI) - pow2(muK);
    double yIKmin = 2*muI*muK;
    double sigmasqua = 1 + pow2(muI) - pow2(muK);
    
    double gdetdimless = gDet(yij,yjk,yik,mui,muj,muk);			
    double gdetdimlessij = (yik*yjk-2.*pow2(muk)*yij)/4.;
    double gdetdimlessjk = (yik*yij-2.*pow2(mui)*yjk)/4.;
    
    double r = 0;
    if ( kineMapType == 3) {
      r =
	(
	 sigmasqua
	 + sqrt( pow2(yIK) - pow2(yIKmin) )
	 *( yjk - yjkmin - ( yij - yijmin ) )
	 /( yjk - yjkmin + ( yij - yijmin ) )
	 )/2.;
    }
    else if ( kineMapType == 4 ) {
      double muijsqua = pow2(mui) + pow2(muj) + yij;
      double lambda = 1 + pow2(muijsqua) + pow4(muk)
	-2*muijsqua - 2*pow2(muk)
	-2*muijsqua*pow2(muk);
      r =
	(
	 sigmasqua
	 + sqrt( pow2(yIK) - pow2(yIKmin) )
	 *( 1 - pow2(mui) - pow2(muj) + pow2(muk) - yij)
	 /sqrt( lambda )
	 )/2.;
    }
    // if the kineMapType is not implemented, complain
    else {
      if (verbose>=3) {
	printOut("VinClu::map3to2massive", "kinematics map type "
		 + num2str(kineMapType) + "not implemented");
      }
      clusterOK = false;
    }
    
    
    double bigsqr =
      sqrt(
	   16.*( r*(1.-r) - (1.-r)*pow2(muI) - r*pow2(muK) )*gdetdimless
	   + ( pow2(yik) - pow2(yikmin) )*( pow2(yIK) - pow2(yIKmin) )
	   );
    
    double x =
      (
       sigmasqua*( pow2(yik) - pow2(yikmin) + 4.*gdetdimlessij)
       + 8.*r*( gdetdimless - gdetdimlessij )
       + bigsqr*( 1. - pow2(mui) - pow2(muj) + pow2(muk) - yij)
       )/(2.*( 4.*gdetdimless + pow2(yik) - pow2(yikmin) ));
    
    double z =
      (
       sigmasqua*( pow2(yik) - pow2(yikmin) + 4.*gdetdimlessjk)
       + 8.*r*( gdetdimless - gdetdimlessjk )
       - bigsqr*( 1. + pow2(mui) - pow2(muj) - pow2(muk) - yjk)
       )/(2.*( 4.*gdetdimless + pow2(yik) - pow2(yikmin) ));
    
    
    // Compute reclustered momenta
    pOutI       =     x*pi +     r*pj +     z*pk;
    pOutK       = (1-x)*pi + (1-r)*pj + (1-z)*pk;
	
    // MR 12.7.2010: check if I got on-shell momenta
    // MR 12.11.2010: normalise the off-shellness to Ecm to get 
    // a dimensionless quantity to compare to TINY
    // PS 23.2.2011: changed to squared quantities 
    double offshellnessI = abs(pOutI.m2Calc() - pow2(mI))/Ecmsqua;
    double offshellnessK = abs(pOutK.m2Calc() - pow2(mK))/Ecmsqua;
    if (offshellnessI > TINY || offshellnessK >TINY) {      
      if (verbose>=3) {
	printOut("VinClu::map3to2massive","on-shell check failed");
      }
      clusterOK = false;
    }
    
    return clusterOK;
  }
  
  //**********************************************************************
  
  // Rambo phase space generator
  // p[0][i] is the energy of the i'th parton (out of N, with ECM^2 = S)
  
  double Rambo::genPoint(double S,int N,double p[4][8])
  {
    double* R=new double[4];for (int j=0;j<4;j++) R[j]=0.0;
    for (int i=0;i<N;i++) {
      double c=-1.0+2.0*rndmPtr->flat();
      double s=sqrt(1.0-c*c);
      double f=2.0*M_PI*rndmPtr->flat();
      double r12=0.0;
      while (r12==0.0) {
	double r1=rndmPtr->flat();
	double r2=rndmPtr->flat();
	r12=r1*r2; }
      double En=-log(r12);
      p[0][i]=En;
      p[3][i]=En*c;
      p[2][i]=En*s*cos(f);
      p[1][i]=En*s*sin(f);
      for (int j=0;j<4;j++) R[j]+=p[j][i];
    }
    double Rmass=sqrt(R[0]*R[0]-R[1]*R[1]-R[2]*R[2]-R[3]*R[3]);
    for (int j=0;j<4;j++) R[j]/=-Rmass;
    double a=1.0/(1.0-R[0]);
    double x=sqrt(S)/Rmass;
    for (int i=0;i<N;i++) {
      double bq=R[1]*p[1][i]+R[2]*p[2][i]+R[3]*p[3][i];
      for (int j=1;j<4;j++)
	p[j][i]=x*(p[j][i]+R[j]*(p[0][i]+a*bq));
      p[0][i]=x*(-R[0]*p[0][i]+bq);
    }
    delete R;
    return 1.0;
  }
  
  // MR 6.2.2011: generalise genPoint according to the original paper
  // The momenta are not distributed flat in phase space anymore,
  // return the weight of the phase space configutation
  
  double Rambo::genPoint(double S,int N,double m[8], double p[4][8])
  {
    double Ecm = sqrt(S);
    // call the massless genPoint, initializing weight
    double weight = genPoint(S,N,p);
    bool massesnonzero = false;
    // set up the function determining the rescaling parameter xi
    vector<double> mvec;
    vector<double> energiesvec;
    for (int i=0; i<N; i++) {
      mvec.push_back(m[i]);
      energiesvec.push_back(p[0][i]);
      if (pow2(m[i]/Ecm) > TINY) {
        massesnonzero = true;
      }
    }
    // if none of the reduced masses is > TINY, return
    if (!massesnonzero) {
      return weight;
    }
    
    TXiFunctor rhs = TXiFunctor(mvec, energiesvec);
    double xi = zbrentAlt(rhs, Ecm, 0., 1., 1e-10);
    // rescale all the momenta
    for (int iMom=0; iMom<N; iMom++) {
      p[1][iMom] *= xi;
      p[2][iMom] *= xi;
      p[3][iMom] *= xi;
      p[0][iMom] = sqrt( pow2(m[iMom]) + pow2(xi)*pow2(p[0][iMom]) );
    }
    // calculate the quantities needed for the calculation of the weight
    double sumpnorm = 0.;
    double prodpnormdivE = 1.;
    double sumpnormsquadivE = 0.;
    for (int iMom=0; iMom<N; iMom++) {
      double pnormsqua = 
        pow2(p[1][iMom])+pow2(p[2][iMom])+pow2(p[3][iMom]);
      double pnorm = sqrt(pnormsqua);
      sumpnorm += pnorm;
      prodpnormdivE *= pnorm/p[0][iMom];
      sumpnormsquadivE += pnormsqua/p[0][iMom];
    }
    // There's a typo in eq. 4.11 of the Rambo paper by Kleiss, 
    // Stirling and Ellis, the Ecm below is not present there
    weight *=  
      pow(sumpnorm/Ecm,2*N-3)*prodpnormdivE*Ecm/sumpnormsquadivE;
    return weight;
  }
  
  // MR 15.3.2011: need a function which produces a sequence of 
  // momenta where the first two particles become collinear for 
  // testing purposes
  // the return vector does NOT have a fixed size, it has a maximal
  // length of nSteps, but it stops if the relative error between 
  // the ratio of two consecutive invariants differs from 
  // rescalingFactor by more than rescalingFactorTolerance again
  // it returns an empty vector if rescalingFactorTolerance was never
  // respected
  vector< vector<Vec4> > Rambo::collinearSequence( 
    double S, int nParticles, int nSteps, int n1, int n2, 
    double rescalingFactor, double rescalingFactorTolerance
    ) {
    if (nParticles <= 2 || nParticles >= 9) {
      return vector< vector<Vec4> >();
    }
    // produce random momenta with rambo
    // demand that the angle between p0 and p1 be between pi/8 and pi/4
    // and that the two energies should not differ by more than an 
    // order of magnitude
    double costheta01 = 0.;
    double energyratio01 = 0.;
    int ntries = 0;
    const int ntriesmax = 1000;
    vector<Vec4> pSav(0);
    while (
      (costheta01 < cos(M_PI/4) || costheta01 > cos(M_PI/8) 
        || energyratio01 < 0.1 || energyratio01 > 10.0)
      && ntries <= ntriesmax
      ) {
      pSav.resize(0);
      double p[4][8];
      this->genPoint(S,nParticles, p);
      for (int iParton=0;iParton<nParticles;iParton++) { 
        Vec4 VecP;
        VecP.px(p[1][iParton]);
        VecP.py(p[2][iParton]);
        VecP.pz(p[3][iParton]);
        VecP.e(p[0][iParton]);
        pSav.push_back(VecP);
      }
      costheta01 = costheta(pSav[0], pSav[1]);
      energyratio01 = pSav[0].e()/pSav[1].e();
      ntries++;
    }
    
    // the second particle becomes collinear to the first
    double p0proj = dot3(pSav[0],pSav[1])/dot3(pSav[0],pSav[0]);
    Vec4 p1perp = pSav[1] - p0proj*pSav[0];
    p1perp.e(0.0);
    vector< vector<Vec4> > collSeq;
    double eCM = sqrt(S);
    bool invRatioOK = false;
    double lastinv = eCM;
    
    for (int iResc = 0; iResc < nSteps; iResc++) {
      vector<Vec4> pResc(pSav);
      double r = pow(rescalingFactor, iResc/2.0);
      pResc[1] = r*p1perp + p0proj*pSav[0];
      pResc[1].e(pResc[1].pAbs());
      pResc[2] += (1-r)*p1perp;
      pResc[2].e(pResc[2].pAbs());
      
      // rescale to adjust the total energy
      double eCMNow = 0.0;
      for (vector<Vec4>::iterator pRescIt = pResc.begin(); pRescIt != pResc.end(); pRescIt++) {
        eCMNow += pRescIt->e();
      }
      for (vector<Vec4>::iterator pRescIt = pResc.begin(); pRescIt != pResc.end(); pRescIt++) {
        (*pRescIt) *= eCM/eCMNow;
      }
      
      double inv = 2*pResc[0]*pResc[1];
      double invRatio = inv/lastinv;
      double rDeviation = abs(1-invRatio/rescalingFactor);
      lastinv = inv;
      if (rDeviation < rescalingFactorTolerance) {
        invRatioOK = true;
      }
      else if (invRatioOK) {
        break;
      }
      
      // make the vectors with indices n1 and n2 become collinear
      int nmin = min(n1, n2);
      int nmax = max(n1, n2);
      // full switch if nmin is not zero or 1
      if (nmin != 0 && nmin != 1) {
        Vec4 p0 = pResc[nmin];
        pResc[nmin] = pResc[0];
        pResc[0]  = p0;
        Vec4 p1 = pResc[nmax];
        pResc[nmax] = pResc[1];
        pResc[1]  = p1;
      }
      // switch nmax with 1 if nmin=0, mnax not 1
      else if (nmin == 0 && nmax != 1) {
        /*
        cout << "nmin = " << nmin << ", nmax = " << nmax << endl;
        cout << "before switch:" << endl;
        cout << "2*p0*p1 = " << scientific << 2*pResc[0]*pResc[1] << endl;
        for (vector<Vec4>::iterator vPtr = pResc.begin(); vPtr != pResc.end(); vPtr++) {
          cout << *vPtr;
        }
        cout << endl;
        */
        Vec4 p1 = pResc[nmax];
        pResc[nmax] = pResc[1];
        pResc[1]  = p1;
        /*
        cout << "after switch:" << endl;
        for (vector<Vec4>::iterator vPtr = pResc.begin(); vPtr != pResc.end(); vPtr++) {
          cout << *vPtr;
        }
        cout << "2*p"<< nmin << "p" << nmax << " = " << 2*pResc[nmin]*pResc[nmax] << endl;
        */
      }
      // only switch nmax with 0 if 1 is one of the positions
      else if (nmin ==1) {
        Vec4 p0 = pResc[nmax];
        pResc[nmax] = pResc[0];
        pResc[0]  = p0;
      }
      collSeq.push_back(pResc);
    }
    if (invRatioOK) {
      return collSeq;   
    }
    else {
      return vector< vector<Vec4> >();
    }

     
    
  }

  // End VINCIA namespace
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  double Rambo::genPointPrime(double ecm, int n, vector<double> masses, vector<Vec4>& retVec){
    Vec4 qtot(0);
    retVec = vector<Vec4> (n, Vec4(0.));
    for (int iMom = 0; iMom < n; iMom++) {
      double c = 2.*rndmPtr->flat()-1.;
      double s = sqrt( 1-pow2(c) );
      double phi = 2*M_PI*rndmPtr->flat();
      double q0 = -log( rndmPtr->flat()*rndmPtr->flat() );
      double qx = q0*s*sin(phi);
      double qy = q0*s*cos(phi);
      double qz = q0*c;
      Vec4 veci;
      veci.p(qx, qy, qz, q0);
      qtot += veci;
      retVec[iMom] = veci;
    }
    // rescale and boost the momenta
    double mqtot = sqrt(qtot*qtot);
    double gamma = qtot.e()/mqtot;
    Vec4 b = -qtot/mqtot;
    double a = 1/(1+gamma);
    double x = ecm/mqtot;
    for (int iMom = 0; iMom < n; iMom++) {
      double bdotqi = dot3(b, retVec[iMom]);
      double qi0 = retVec[iMom].e();
      retVec[iMom].e( x*(gamma*qi0 + bdotqi) );
      retVec[iMom].px( x*( retVec[iMom].px() + b.px()*qi0 + a*bdotqi*b.px() ) );
      retVec[iMom].py( x*( retVec[iMom].py() + b.py()*qi0 + a*bdotqi*b.py() ) );
      retVec[iMom].pz( x*( retVec[iMom].pz() + b.pz()*qi0 + a*bdotqi*b.pz() ) );
    }
    // massless particles are now OK
    if (masses.size() == 0) {
      return 1.;
    }
    double massmax = 0.;
    for (vector<double>::iterator massPtr = masses.begin(); massPtr < masses.end(); massPtr++) {
      massmax = max(massmax, masses[*massPtr]);
    }
    if (massmax == 0.) {
      return 1.;
    }
    // determine xi
    vector<double> emasslessVec;
    for (vector<Vec4>::iterator mPtr = retVec.begin(); mPtr != retVec.end(); mPtr++) {
      emasslessVec.push_back( mPtr->e() );
    }
    vector<double> mVec = masses;
    while (int(mVec.size()) < n) {
      mVec.push_back(0.);
    }
    
    TXiFunctor rhs = TXiFunctor(mVec, emasslessVec);
    double xi = zbrentAlt(rhs, ecm, 0., 1., 1e-10);
    
    double weight = pow(xi, 2*n-3);
    double ksquasum = 0.;
    
    for (int iMom = 0; iMom != n; iMom++) {
      // put on shell
      retVec[iMom].e( sqrt( pow2(mVec[iMom]) + pow2(xi*retVec[iMom].e()) )  );
      // rescale
      retVec[iMom].px( xi*retVec[iMom].px() );
      retVec[iMom].py( xi*retVec[iMom].py() );
      retVec[iMom].pz( xi*retVec[iMom].pz() );
      // weight
      weight *= retVec[iMom].pAbs()/retVec[iMom].e();
      ksquasum += retVec[iMom].pAbs2()/retVec[iMom].e()/ecm;
    }
    weight /= ksquasum;
    
    return weight;
  }

}

