// vincia03-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study light-flavor fragmentation 
// at LEP (compared to L3 udsc measurements), using VINCIA's ROOT-based 
// runtime display. 
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2011 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot.h"
// FASTJET
// To use FastJet, do a search on FastJet below and uncomment as necessary
// #include "fastjet/PseudoJet.hh"
// #include "fastjet/ClusterSequence.hh"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects. 
  Pythia pythia; 
  VinciaPlugin vincia(&pythia,"vincia03.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event* event = &pythia.event;

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  

  // Extract settings to be used in the main program.  
  double eCM              = settingsPtr->parm("Beams:eCM");
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");
  int    nShow            = settingsPtr->mode("Main:timesToShow");

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  string antennaFile      = settingsPtr->word("Vincia:antennaFile");
  bool   uncertaintyBands = vinciaOn && settingsPtr->flag("Vincia:uncertaintyBands");
  int    verbose          = settingsPtr->mode("Vincia:verbose");
  bool   sectorShowers    = settingsPtr->flag("VinciaSector");

  // Extract matching settings
  int    maxMatchTree     = settingsPtr->mode("VinciaMatching:order");
  bool   matching         = maxMatchTree >= 1;

  //************************************************************************
  // VinciaRoot
  //************************************************************************

  // Create VinciaRoot object 
  VinciaRoot vroot; 

  // Initialize some ROOT displays (name,x-pixels,y-pixels,x-offset,y-offset)
  // Book additional displays here as you like.
  vroot.createDisplay("alphaS",330,440);
  vroot.createDisplay("Nch",330,440);
  vroot.createDisplay("Npartons",330,440);
  vroot.createDisplay("Nquarks",330,440);
  vroot.createDisplay("Mesons",330,440);
  vroot.createDisplay("Baryons",330,440);
  vroot.createDisplay("Kaons",330,440);
  vroot.createDisplay("protons",330,440);
  vroot.createDisplay("Lambdas",330,440);
  vroot.createDisplay("Xis",330,440);
  vroot.createDisplay("x",330,440);
  vroot.createDisplay("Thrust",330,440);
  vroot.createDisplay("C",330,440);
  vroot.createDisplay("D",330,440);
  vroot.createDisplay("BW",330,440);
  vroot.createDisplay("BT",330,440);

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose(verbose);

  // Switch stats on/off (in ROOT conventions, but with 0 = off)
  //vroot.setOptStat(0);

  // Set the name and version of the MCs we are running 
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  vroot.addRef("Vincia "+vcVer);
  if (vinciaOn && matching) vroot.addRef("MadGraph 4.426");
  vroot.addRef("Pythia "+pyVer);

  // Set label for plot legend
  string mcLabel = "Vincia";  
  if (!vinciaOn) mcLabel = "Pythia";
  else if (sectorShowers) mcLabel = "Vincia (sector)";
  vroot.setLabel(mcLabel);

  // Set condensed MC name for saving to dat files
  string mcFile = "vc";  
  if (!vinciaOn) mcFile = "py";
  else if (sectorShowers) mcFile = "vs";    
  if (!settingsPtr->flag("HadronLevel:all")) mcFile+="-nohad";

  // Optionally set name for saved dat file(s) to read in and overplot
  // Compare Global Vincia to Pythia
  // Compare Sector Vincia to global Vincia
  string mcOld = "";    
  if (vinciaOn) {
    mcOld = "py";
    if ( sectorShowers ) mcOld = "vc";
  }
  if (!settingsPtr->flag("HadronLevel:all")) mcOld+="-nohad";

  // Set names for uncertainty bands, if using them (for plot legends)
  if (uncertaintyBands) {
    vroot.addUncertainty("Def",0,1);
    vroot.addUncertainty("#muR",2,3);
    vroot.addUncertainty("Finite",4,5);
    vroot.addUncertainty("QMatch",6,7);
    vroot.addUncertainty("Ord",8,9);
    vroot.addUncertainty("1/N_{C}^{2}",10,11);
  }

  // Use a different color scheme for PYTHIA plots
  if ( !vinciaOn ) vroot.setColorScheme(3);
  // Use a different color scheme if uncertainty bands off
  else {
    if ( mcOld.find("vs") != string::npos ) vroot.setColorScheme(2);
  }

  // Set first update and save event
  double iSave   = 1000;
  int    iUpdate = 100;

  //************************************************************************
  // Initialize PYTHIA (automatically initializes VINCIA) 
  //************************************************************************

  // Initialize
  pythia.init( 11, -11, eCM);

  // List changed settings
  pythia.settings.listChanged();

  // Check that Z0 decay channels set correctly.
  pythia.particleData.listChanged();
 
  //***************************************************************************
  // ROOT
  //***************************************************************************

  // Form name for ROOT file
  ostringstream outfile;
  // Put root files in subdirectory defined at beginning of file
  if (not vinciaOn) {
    outfile<<"py03";
  } else {
    outfile<<"vc03";
  }
  if (!settingsPtr->flag("HadronLevel:all")) outfile<<"-nohad";
  outfile<<".root";

  // 
  // DA
  // 

  //****************
  // Book Histograms

  // Vincia uncertainty bands: number of variations expected 
  int nVariations = max(1,vincia.nVariations());

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi;

  // Book alphaS histogram
  nBins = 200;
  xLo   = 0.;
  xHi   = 100.;
  TH1F *alphaS=new TH1F("alphaS","#alpha_{s}(Q)",nBins,xLo,xHi);
  // Set properties 
  alphaS->SetMinimum(0.0);
  alphaS->SetMaximum(0.5);
  alphaS->GetXaxis()->SetTitle("Q/GeV");
  // Add to Vincia's list for run-time display and post-run saving
  // logy = false, normalize=false, showStatErr=false
  vroot.addTH1FPtr(alphaS,"alphaS",false,false,false);

  // Fill histogram
  for ( int iBin=0; iBin<nBins; ++iBin ) {
    double Q = xLo + (iBin+0.5)*(xHi-xLo)/nBins;
    double as   = vincia.shower.alphaS.alphaS(pow2(Q));
    if (as < alphaS->GetMaximum()) alphaS->Fill(Q,as);
    else alphaS->Fill(Q,alphaS->GetMaximum());
    alphaS->SetBinError(iBin+1,0.);	
  }

  // Event Weights
  nBins = 200;
  xLo   = -2.5;
  xHi   = 2.5;
  TH1F *HnPN=new TH1F("Weights","Event Weights",nBins,xLo,xHi);
  //  vroot.addTH1FPtr(HnPN,"MyPlots",true,false,false);
  HnPN->SetFillColor(3); 

  // nPartons
  TH1F *HnPartons=new TH1F("nPartons","N_{Partons}(Q_{had})", 50, -0.5, 49.5);
  vroot.addTH1FPtr(HnPartons,"Npartons",true,true,true);
  HnPartons->SetMinimum(0.0005);
  HnPartons->SetMaximum(0.5);

  // Quark Pairs
  TH1F *HnQuarks=new TH1F("nQuarks","N_{Quarks}(Q_{had})", 10, -0.5, 9.5);
  vroot.addTH1FPtr(HnQuarks,"Nquarks",true,true,true);
  HnQuarks->SetMinimum(0.00005);
  HnQuarks->SetMaximum(5.0);

  // Thrust (non-B flavor)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo = 0.0;
  xHi = 0.5;
  vector<TH1F*> VThrustLight;
  VThrustLight.push_back(new TH1F("ThrustLight","1-Thrust (udsc)",nBins,xLo,xHi));
  double wThrust = nBins/(xHi-xLo);
  VThrustLight[0]->GetXaxis()->SetTitle("1-T (udsc)");
  VThrustLight[0]->GetYaxis()->SetTitle("1/N dN/d(1-T)");
  VThrustLight[0]->SetMinimum(0.0005);
  VThrustLight[0]->SetMaximum(90.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VThrustLight.push_back(new TH1F(string("ThrustLight"+num2str(iVar,0)).c_str(),
				    " ",
				    nBins,xLo,xHi));  
  vroot.addTH1FPtr(VThrustLight,"Thrust",true,true,true,"data/L3-91-Thrust-udsc.vec",mcOld+"-T-udsc.dat");

  // C (light)
  nBins = 20; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 1.0;
  vector<TH1F*> VCLight;
  VCLight.push_back(new TH1F("CLight","C Parameter (udsc)",nBins,xLo,xHi));
  double wC = nBins/(xHi-xLo);
  VCLight[0]->GetXaxis()->SetTitle("C (udsc)");
  VCLight[0]->GetYaxis()->SetTitle("1/N dN/dC");
  VCLight[0]->SetMinimum(0.0005);
  VCLight[0]->SetMaximum(90.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VCLight.push_back(new TH1F(string("CLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VCLight,"C",true,true,true,"data/L3-91-C-udsc.vec",
		   mcOld+"-C-udsc.dat");
  
  // D (udsc)
  nBins = 25; // Important = multiplum of # of data bins, for ratio rebinning
  xLo   = 0.0;
  xHi   = 0.8;
  vector<TH1F*> VDLight;
  VDLight.push_back(new TH1F("DLight","D Parameter (udsc)",nBins,xLo,xHi));
  double wD = nBins*1.0/(xHi-xLo);
  VDLight[0]->GetXaxis()->SetTitle("D (udsc)");
  VDLight[0]->GetYaxis()->SetTitle("1/N dN/dD");
  VDLight[0]->SetMinimum(0.0005);
  VDLight[0]->SetMaximum(90.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VDLight.push_back(new TH1F(string("DLight"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VDLight,"D",true,true,true,"data/L3-91-D-udsc.vec",
		   mcOld+"-D-udsc.dat");

  // nCharged (udsc)
  nBins = 30;
  xLo   = -1.0;
  xHi   = 59.0;
  vector<TH1F*> VnCharged;
  VnCharged.push_back(new TH1F("nCh","Charged Multiplicity (udsc)",nBins,xLo,xHi));
  VnCharged[0]->GetXaxis()->SetTitle("N_{ch} (udsc)");
  VnCharged[0]->GetYaxis()->SetTitle("Probability(N_{ch})");
  VnCharged[0]->SetMinimum(0.000005);
  VnCharged[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VnCharged.push_back(new TH1F(string("nCh"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Only overplot PYTHIA when running VINCIA
  if (vinciaOn) {
    vroot.addTH1FPtr(VnCharged,"Nch",true,true,true,"data/L3-91-Nch-udsc.vec",
		     mcOld+"-Nch-udsc.dat");
  } else {
    vroot.addTH1FPtr(VnCharged,"Nch",true,true,true,"data/L3-91-Nch-udsc.vec");
  }

  // Mesons
  nBins = 10;
  xLo   = -0.5;
  xHi   =  9.5;
  vector<TH1F*> VMesons;
  VMesons.push_back(new TH1F("Mesons","Meson Fractions",10, -0.5, 9.5));
  VMesons[0]->GetXaxis()->SetBinLabel(1,"#pi^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(2,"#pi^{0}");
  VMesons[0]->GetXaxis()->SetBinLabel(3,"K^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(4,"#eta");
  VMesons[0]->GetXaxis()->SetBinLabel(5,"#eta'");
  VMesons[0]->GetXaxis()->SetBinLabel(6,"#rho^{#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(7,"#rho^{0}"); 
  VMesons[0]->GetXaxis()->SetBinLabel(8,"K^{*#pm}");
  VMesons[0]->GetXaxis()->SetBinLabel(9,"#omega");
  VMesons[0]->GetXaxis()->SetBinLabel(10,"#phi");
  vroot.addTH1FPtr(VMesons,"Mesons",true,false,true,"data/LEP91-mesons.vec");
  VMesons[0]->SetMaximum(20.01); 
  VMesons[0]->SetMinimum(0.0002); 

  // Baryons
  nBins = 11;
  xLo   = -0.5;
  xHi   = 10.5;
  vector<TH1F*> VBaryons;
  VBaryons.push_back(new TH1F("Baryons","Baryon Fractions",nBins, xLo, xHi));  
  VBaryons[0]->GetXaxis()->SetBinLabel(1,"p");
  VBaryons[0]->GetXaxis()->SetBinLabel(2,"#Lambda");
  VBaryons[0]->GetXaxis()->SetBinLabel(3,"#Lambda/p");
  VBaryons[0]->GetXaxis()->SetBinLabel(4,"#Lambda/K");
  VBaryons[0]->GetXaxis()->SetBinLabel(5,"#Sigma^{#pm}");
  VBaryons[0]->GetXaxis()->SetBinLabel(6,"#Sigma^{0}");
  VBaryons[0]->GetXaxis()->SetBinLabel(7,"#Delta");
  VBaryons[0]->GetXaxis()->SetBinLabel(8,"#Sigma^{*}");
  VBaryons[0]->GetXaxis()->SetBinLabel(9,"#Xi^{#pm}");
  VBaryons[0]->GetXaxis()->SetBinLabel(10,"#Xi^{*0}");
  VBaryons[0]->GetXaxis()->SetBinLabel(11,"#Omega");
  vroot.addTH1FPtr(VBaryons,"Baryons",true,false,true,"data/LEP91-baryons.vec");
  VBaryons[0]->SetMaximum(2.01); 
  VBaryons[0]->SetMinimum(0.00002); 

  // Ln(1/x) (udsc)
  nBins = 43;
  xLo   = -0.2;
  xHi   = 8.4;
  vector<TH1F*> Vx;
  Vx.push_back(new TH1F("LnXInv","Ln(1/x) (udsc)",nBins,xLo,xHi));
  double wx = nBins/(xHi-xLo);
  Vx[0]->GetXaxis()->SetTitle("Ln(1/x) (udsc)");
  Vx[0]->GetYaxis()->SetTitle("1/N_{ch} dN_{ch}/dLn(1/x)");
  Vx[0]->SetMinimum(0.0005);
  Vx[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vx.push_back(new TH1F(string("LnXInv"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(Vx,"x",true,true,true,"data/L3-91-LogInvx-udsc.vec",
		   mcOld+"-x-udsc.dat");

  // x(K+-)
  nBins = 250;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 0.5;
  vector<TH1F*> VxKch;
  VxKch.push_back(new TH1F("xKch","x(K^{#pm})",nBins,xLo,xHi));
  double wxKch = nBins/(xHi-xLo);
  VxKch[0]->GetXaxis()->SetTitle("x_{p}");
  VxKch[0]->GetYaxis()->SetTitle("1/N_{K} dN_{K}/dx");
  VxKch[0]->SetMinimum(0.05);
  VxKch[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VxKch.push_back(new TH1F(string("xKch"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VxKch,"Kaons",true,true,true,"data/SLD-91-xKch.vec",
		   mcOld+"-xKch.dat");

  // x(protons)
  nBins = 500;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 1.0;
  vector<TH1F*> Vxp;
  Vxp.push_back(new TH1F("xp","x(protons)",nBins,xLo,xHi));
  double wxp = nBins/(xHi-xLo);
  Vxp[0]->GetXaxis()->SetTitle("x_{p}");
  Vxp[0]->GetYaxis()->SetTitle("1/N_{p} dN_{p}/dx");
  Vxp[0]->SetMinimum(0.0005);
  Vxp[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    Vxp.push_back(new TH1F(string("xp"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(Vxp,"protons",true,true,true,"data/SLD-91-xp.vec",
		   mcOld+"-xp.dat");

  // x(Lambda)
  nBins = 500;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 1.0;
  vector<TH1F*> VxLam;
  VxLam.push_back(new TH1F("xLam","x(Lambda)",nBins,xLo,xHi));
  double wxLam = nBins/(xHi-xLo);
  VxLam[0]->GetXaxis()->SetTitle("x_{p}");
  VxLam[0]->GetYaxis()->SetTitle("1/N_{#Lambda} dN_{#Lambda}/dx");
  VxLam[0]->SetMinimum(0.005);
  VxLam[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VxLam.push_back(new TH1F(string("xLam"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VxLam,"Lambdas",true,true,true,"data/Aleph-91-xL.vec",
		   mcOld+"-xLam.dat"); 

  // x(Xi)
  nBins = 50;  // (needed due to non-uniform bins in data)
  xLo   = 0;
  xHi   = 0.50;
  vector<TH1F*> VxXi;
  VxXi.push_back(new TH1F("xXi","x(#Xi^{#pm})",nBins,xLo,xHi));
  double wxXi = nBins/(xHi-xLo);
  VxXi[0]->GetXaxis()->SetTitle("x_{E}");
  VxXi[0]->GetYaxis()->SetTitle("1/N_{#Xi} dN_{#Xi}/dx");
  VxXi[0]->SetMinimum(0.05);
  VxXi[0]->SetMaximum(50.);
  for (int iVar=1; iVar<nVariations; iVar++) 
    VxXi.push_back(new TH1F(string("xXi"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VxXi,"Xis",true,true,true,"data/Aleph-91-xXi.vec",
		   mcOld+"-xXi.dat");

  // Wide Jet Broadening
  nBins = 20; // Must match or be multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.3;
  vector<TH1F*> VBjWide;
  VBjWide.push_back(new TH1F("BWide","Wide Jet Broadening (udsc)",nBins,xLo,xHi));
  double wBjWide = nBins/(xHi-xLo);
  VBjWide[0]->GetXaxis()->SetTitle("B_{W} (udsc)");
  VBjWide[0]->GetYaxis()->SetTitle("1/N dN/dB_{W}");
  VBjWide[0]->SetMaximum(500.0); 
  VBjWide[0]->SetMinimum(0.0005);   
  for (int iVar=1; iVar<nVariations; iVar++) 
    VBjWide.push_back(new TH1F(string("BWide"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VBjWide,"BW",true,true,true,"data/L3-91-BWide-udsc.vec",
		   mcOld+"-BW-udsc.dat");

  // Total Jet Broadening
  nBins = 20; // Must match or be multiplum of # bins in data, for ratio rebinning
  xLo = 0.0;
  xHi = 0.4;
  vector<TH1F*> VBjTot;
  VBjTot.push_back(new TH1F("BTot","Total Jet Broadening (udsc)",20, 0.0, 0.4));
  double wBjTot = 200.0/0.4;
  VBjTot[0]->GetXaxis()->SetTitle("B_{T} (udsc)");
  VBjTot[0]->GetYaxis()->SetTitle("1/N dN/dB_{T}");
  VBjTot[0]->SetMaximum(500.0); 
  VBjTot[0]->SetMinimum(0.0005);     
  for (int iVar=1; iVar<nVariations; iVar++) 
    VBjTot.push_back(new TH1F(string("BTot"+num2str(iVar,0)).c_str(),
			       " ",nBins,xLo,xHi));
  vroot.addTH1FPtr(VBjTot,"BT",true,true,true,"data/L3-91-BTot-udsc.vec",
		   mcOld+"-BT-udsc.dat");

  // Event analysis tools
  Thrust Thrust(1);
  Sphericity SphLin(1,2);

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Number of quarks, partons, and charged particles
  double nQsum=0.0;
  double nPsum=0.0;
  double nCsum=0.0;
  double nMesSum[20] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};
  double nBarSum[12] = {0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.};

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    // Verbose output
    if (iEvent%(max(1,nEvent/max(nShow,1))) == 0)  
      cout << " Now begin event " << iEvent << endl;

    // Generate next event
    if (!pythia.next()) continue;
    event = &pythia.event;

    // List the first few events
    if (iEvent < 3) event->list();

    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;

    // Bool to reject events with B's
    bool isB = false;
    for (int i=1;i<event->size();i++) 
      if (abs((*event)[i].id()) == 5) isB = true;  
 
    // Main loop over event record
    for (int i=1;i<event->size();i++) {

      // Count up final-state charged hadrons
      double x = 2.0*(*event)[i].pAbs()/eCM;
      double xE = 2.0*(*event)[i].e()/eCM;
      if ((*event)[i].isCharged() && (*event)[i].status() > 80) {
	nc++;
	double LnXInv=log(1./x);
	if (! isB) {
	  Vx[0]->Fill(LnXInv,wx);
	  if (uncertaintyBands) {
	    for (int iVar=1; iVar<nVariations; iVar++) {
	      if (vincia.weight(iVar) > 0.0) {
		Vx[iVar]->Fill(LnXInv,wx*vincia.weight(iVar));
	      }
	    }      	  
	  }
	}
      }
      // Find last parton-level partons
      int iDau1 = (*event)[i].daughter1();
      if (iDau1 == 0 || abs((*event)[iDau1].status()) > 80) {
	// Count up partons and quarks
	if ((*event)[i].isQuark() || (*event)[i].isGluon() ) np++;
	if ((*event)[i].isQuark()) nq++;	
      }
      // Find identified particles
      /* HMesons:
	# 211 0: e+ e- --> pi+- X
	# 111 1: e+ e- --> pi0 X
	# 321 2: e+ e- --> K+- X
	# 221 3: e+ e- --> eta X
	# 331 4: e+ e- --> etaprime(958) X
	# 213 5: e+ e- --> rho(770)+- X
	# 113 6: e+ e- --> rho(770)0 X
	# 323 7: e+ e- --> K*(892)+- X
	# 223 8: e+ e- --> omega(782) X
	# 333 9: e+ e- --> phi(1020) X
	# 10111 10: e+ e- --> a0(980) X
	# 10221 11: e+ e- --> f0(980) X
      */
      /* HBaryons:
	 # 2212      0 : e+ e- --> p (pbar) X
	 # 3122      1: e+ e- --> Lambda (Lambdabar) X
	 # 3122/2212 2: e+ e- --> Lambda (Lambdabar) / p(pbar)
	 # 3122/321  3: e+ e- --> Lambda (Lambdabar) / K+-
	 # 3222+3112 4: e+ e- --> Sigma+- (Sigmabar+-) X
	 # 3212      5: e+ e- --> Sigma0 (Sigmabar0) X
	 # 2224      6 : e+ e- --> Delta(1232)++ (Delta(1232)--) X
	 # 3224+3114 7: e+ e- --> Sigma(1385)+- (Sigmabar(1385)+-) X
	 # 3312      8: e+ e- --> Xi- (Xibar+) X
	 # 3324      9: e+ e- --> Xi(1530)0 (Xibar(1530)0) X
	 # 3334     10: e+ e- --> Omega- (Omegabar+) X 
      */
      /* HHMesons:
	 # 411    0:  e+ e- --> D+- X
         # 421    1: e+ e- --> D0 (Dbar0) X
         # 413    2: e+ e- --> D*(2010)+- X
         # 443    3: e+ e- --> J/psi X
         # 20443  4: e+ e- --> chi(c1) X
         # 100443 5: e+ e- --> psi(3685) X  	
      */
      int idp = abs((*event)[i].id());
      if (idp == 211) nMesSum[0]++;
      if (idp == 111) nMesSum[1]++;
      if (idp == 321) {
	nMesSum[2]++;
	if ( (x >= 0.016 && x <= 0.164) 
	     || (x >= 0.208 && x <= 0.768) ) VxKch[0]->Fill(x,wxKch);
      }
      if (idp == 221) nMesSum[3]++;
      if (idp == 331) nMesSum[4]++;
      if (idp == 213) nMesSum[5]++;
      if (idp == 113) nMesSum[6]++;
      if (idp == 323) nMesSum[7]++;
      if (idp == 223) nMesSum[8]++;
      if (idp == 333) nMesSum[9]++;
      if (idp == 10111) nMesSum[10]++;
      if (idp == 10221) nMesSum[11]++;
      if (idp == 2212) {
	nBarSum[0]++;
	// Note the SLD data goes to 0.987, but since we "only" use a bin
	// width of 0.002, we have to go to 0.988. The error only concerns
	// the total normalization and should be vanishingly small.
	if ( (x >= 0.016 && x<=0.164 ) 
	     || (x >= 0.230 && 0.988) ) Vxp[0]->Fill(x,wxp);
      }
      if (idp == 3122) {
	nBarSum[1]++;
	if (x >= 0.012 && x <= 0.9) VxLam[0]->Fill(x,wxLam);
      }
      if (idp == 3222) nBarSum[4]++; 
      if (idp == 3112) nBarSum[4]++; 
      if (idp == 3212) nBarSum[5]++; 
      if (idp == 2224) nBarSum[6]++; 
      if (idp == 3114) nBarSum[7]++; 
      if (idp == 3224) nBarSum[7]++; 
      if (idp == 3312) {
	nBarSum[8]++; 
	if (xE >= 0.03 && xE <= 0.5) VxXi[0]->Fill(xE,wxXi);
      }
      if (idp == 3324) nBarSum[9]++; 
      if (idp == 3334) nBarSum[10]++; 
    }

    // Fill Histograms
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;
    HnQuarks->Fill(0.5*nq,1.);
    HnPartons->Fill(1.0*np,1.);    
    if (! isB) {
      VnCharged[0]->Fill(1.0*nc,1.);
      if (uncertaintyBands) {
	for (int iVar=1; iVar<nVariations; iVar++) {
	  if (vincia.weight(iVar) > 0.0) {
	    VnCharged[iVar]->Fill(1.0*nc,vincia.weight(iVar));
	  }
	}      
      }
    }

    // Fill PID histograms
    if (nCsum > 0) {
      VMesons[0]->Reset();
      for (int iMes=0; iMes<=9; iMes++) {
	VMesons[0]->Fill(double(iMes),nMesSum[iMes]/nCsum);
	double err = sqrt(nMesSum[iMes]) / nCsum;
	VMesons[0]->SetBinError(double(iMes)+1,err);
      }
      VBaryons[0]->Reset();
      for (int iBar=0; iBar<=10; iBar++) {
	VBaryons[0]->Fill(double(iBar),nBarSum[iBar]/nCsum);
	double err = sqrt(nBarSum[iBar]) / nCsum;
	VBaryons[0]->SetBinError(double(iBar)+1,err);
      }
      // Fill Lambda/p and Lambda/K
      VBaryons[0]->Fill(double(2),nBarSum[1]/nBarSum[0]);
      double err = sqrt(nBarSum[1]) / nBarSum[0];
      VBaryons[0]->SetBinError(double(2)+1,err);
      VBaryons[0]->Fill(double(3),nBarSum[1]/nMesSum[2]);
      err = sqrt(nBarSum[1]) / nMesSum[2];
      VBaryons[0]->SetBinError(double(3)+1,err);
    }

    // Thrust
    Thrust.analyze( *event );
    // Force light quarks only
    if (! isB) {
      VThrustLight[0]->Fill(1-Thrust.thrust(),wThrust);
      // Uncertainty bands
      if (uncertaintyBands) {
	for (int iVar=1; iVar<nVariations; iVar++) {
	  if (vincia.weight(iVar) > 0.0) {
	    VThrustLight[iVar]->Fill(1-Thrust.thrust(),wThrust*vincia.weight(iVar));
	  }
	}
      }
    }
    
    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( *event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
		    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
		    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
	*SphLin.eigenValue(3);
      // Light quarks
      if (! isB) {
	VCLight[0]->Fill(evC,wC);
	VDLight[0]->Fill(evD,wD);
	// Add uncertainty bands if switched on.
	if (uncertaintyBands) {
	  for (int iVar=1; iVar<nVariations; iVar++) {
	    if (vincia.weight(iVar) > 0.0 && iVar != 0) {
	      VCLight[iVar]->Fill(evC,wC*vincia.weight(iVar));
	      VDLight[iVar]->Fill(evD,wD*vincia.weight(iVar));
	    }
	  }
	}
      }
    }


    // Histogram jet masses
    Vec4 thrustAxis = Thrust.eventAxis(1);   
    Vec4 majorAxis = Thrust.eventAxis(2);   
    Vec4 minorAxis = Thrust.eventAxis(3);   
    Vec4 pPlus, pMinus;
    double bPlus(0.), bMinus(0.), pPlusAbs(0.), pMinusAbs(0.);
    // Loop over event to compute masses
    for (int i=0; i<event->size(); i++) {
      if (not (*event)[i].isFinal() ) continue;
      int id = (*event)[i].id();
      if (abs(id) == 12 || abs(id) == 14 || abs(id) == 16) continue;
      double dotp = (*event)[i].px()*thrustAxis.px()
	+ (*event)[i].py()*thrustAxis.py()
	+ (*event)[i].pz()*thrustAxis.pz();
      double pAbs = (*event)[i].pAbs();
      double cosT = dotp/pAbs/thrustAxis.pAbs();
      double sinT = sqrt(1.-pow2(cosT));
      if (dotp > 0) {
	pPlus     += (*event)[i].p();
	bPlus     += abs(pAbs * sinT);
	pPlusAbs  += pAbs;
      } else {
	pMinus += (*event)[i].p();
	bMinus    += abs(pAbs * sinT);
	pMinusAbs += pAbs;
      }
    }
    bPlus  /= 2.*(pPlusAbs+pMinusAbs);
    bMinus /= 2.*(pMinusAbs+pPlusAbs);    
    if (! isB) {
      VBjWide[0]->Fill(max(bPlus,bMinus),wBjWide);
      VBjTot[0]->Fill(bPlus+bMinus,wBjTot);   
      if (uncertaintyBands) {
	for (int iVar=1; iVar<nVariations; iVar++) {
	  if (vincia.weight(iVar) > 0.0 && iVar != 0) {
	    VBjWide[iVar]->Fill(max(bPlus,bMinus),wBjWide*vincia.weight(iVar));
	    VBjTot[iVar]->Fill(bPlus+bMinus,wBjTot*vincia.weight(iVar));
	  }
	}
      }
    }

    // Periodically update displays
    if ( iEvent == int(iSave) || iEvent == iUpdate ) {
      iUpdate += iSave/sqrt(2.)/5.;
      vroot.updateDisplay();       
      cout<<"vincia03-root: updating plots at event "<<iEvent<<endl;
    }
    // Periodically save displays
    if (iEvent == int(iSave) || iEvent == nEvent-1) {
      iSave *= sqrt(2.0);
      iUpdate = iEvent + (iSave-iEvent)/5.;
      cout<<"vincia03-root: saving plots at event "<<int(iEvent);
      cout<<" next save at event "<<int(iSave)<<endl;

      // Save displays with base name mcFile and file type .eps
      vroot.saveDisplay("Thrust",mcFile,"eps");
      vroot.saveDisplay("D",mcFile,"eps");
      vroot.saveDisplay("C",mcFile,"eps");
      vroot.saveDisplay("Nch",mcFile,"eps");
      vroot.saveDisplay("x",mcFile,"eps");
      vroot.saveDisplay("Kaons",mcFile,"eps");
      vroot.saveDisplay("Lambdas",mcFile,"eps");
      vroot.saveDisplay("protons",mcFile,"eps");
      vroot.saveDisplay("Xis",mcFile,"eps");
      vroot.saveDisplay("Baryons",mcFile,"eps");
      vroot.saveDisplay("Mesons",mcFile,"eps");
      vroot.saveDisplay("BW",mcFile,"eps");
      vroot.saveDisplay("BT",mcFile,"eps");

      // Save histograms as dat files, e.g., for overplotting later
      vroot.saveHistogram(VCLight[0],mcFile,"C-udsc","dat");
      vroot.saveHistogram(VDLight[0],mcFile,"D-udsc","dat");
      vroot.saveHistogram(Vx[0],mcFile,"x-udsc","dat");
      vroot.saveHistogram(VxKch[0],mcFile,"xKch","dat");
      vroot.saveHistogram(VxLam[0],mcFile,"xLam","dat");
      vroot.saveHistogram(Vxp[0],mcFile,"xp","dat");
      vroot.saveHistogram(VxXi[0],mcFile,"xXi","dat");
      vroot.saveHistogram(VnCharged[0],mcFile,"Nch-udsc","dat");
      vroot.saveHistogram(VMesons[0],mcFile,"Mesons","dat");
      vroot.saveHistogram(VBaryons[0],mcFile,"Baryons","dat");
      vroot.saveHistogram(VBjWide[0],mcFile,"BW-udsc","dat");
      vroot.saveHistogram(VBjTot[0],mcFile,"BT-udsc","dat");
      vroot.saveHistogram(VThrustLight[0],mcFile,"T-udsc","dat");
    }
  }

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 2) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.statistics();
  if (vinciaOn) vincia.printInfo();

  // Open ROOT file
  cout<<"Saving root file... "<<outfile.str()<<endl;
  TFile f(outfile.str().c_str(),"recreate");
  // Opened root file:
  cout <<"vincia01kine : opened ROOT file "<<outfile.str()<<endl;  
  alphaS->Write();
  HnPN->Write();
  HnQuarks->Write();
  HnPartons->Write();
  VnCharged[0]->Write();
  Vx[0]->Write();
  VThrustLight[0]->Write();
  VCLight[0]->Write();
  VDLight[0]->Write();
  VBjWide[0]->Write();
  VBjTot[0]->Write();
  // Save variation histograms if they were made
  if (uncertaintyBands) {
    for (int iVar=1; iVar < nVariations; iVar++) {
      VnCharged[iVar]->Write();
      VThrustLight[iVar]->Write();
      VCLight[iVar]->Write();
      VDLight[iVar]->Write();      
      Vx[iVar]->Write();
      VBjTot[iVar]->Write();
      VBjWide[iVar]->Write();
    }
  }
  f.Close();

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
